"""
This is a automatically generated file from the 2018 NIST fundamental constants.
Title: NIST-CODATA Internationally Recommended 2018 Values of the Fundamental Physical Constants - SRD 121
Date: 2019-05-20
DOI: 
URL: https://physics.nist.gov/cuu/Constants/Table/allascii.txt
Access Date: 2020-01-16 23:03:16.193457 UTC

File Authors: QCElemental Authors
"""


# fmt: off
nist_2018_codata = {
    "title": "NIST-CODATA Internationally Recommended 2018 Values of the Fundamental Physical Constants - SRD 121",
    "date": "2019-05-20",
    "doi": "",
    "url": "https://physics.nist.gov/cuu/Constants/Table/allascii.txt",
    "access_data": "2020-01-16 23:03:16.193457",
    "constants": {
        "alpha particle-electron mass ratio": {
            "quantity": "alpha particle-electron mass ratio",
            "unit": "",
            "value": "7294.29954142",
            "uncertainty": "0.000 000 24",
        },
        "alpha particle mass": {
            "quantity": "alpha particle mass",
            "unit": "kg",
            "value": "6.6446573357e-27",
            "uncertainty": "0.000 000 0020 e-27",
        },
        "alpha particle mass energy equivalent": {
            "quantity": "alpha particle mass energy equivalent",
            "unit": "J",
            "value": "5.9719201914e-10",
            "uncertainty": "0.000 000 0018 e-10",
        },
        "alpha particle mass energy equivalent in mev": {
            "quantity": "alpha particle mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "3727.3794066",
            "uncertainty": "0.000 0011",
        },
        "alpha particle mass in u": {
            "quantity": "alpha particle mass in u",
            "unit": "u",
            "value": "4.001506179127",
            "uncertainty": "0.000 000 000 063",
        },
        "alpha particle molar mass": {
            "quantity": "alpha particle molar mass",
            "unit": "kg mol^{-1}",
            "value": "4.0015061777e-3",
            "uncertainty": "0.000 000 0012 e-3",
        },
        "alpha particle-proton mass ratio": {
            "quantity": "alpha particle-proton mass ratio",
            "unit": "",
            "value": "3.97259969009",
            "uncertainty": "0.000 000 000 22",
        },
        "alpha particle relative atomic mass": {
            "quantity": "alpha particle relative atomic mass",
            "unit": "",
            "value": "4.001506179127",
            "uncertainty": "0.000 000 000 063",
        },
        "angstrom star": {
            "quantity": "Angstrom star",
            "unit": "m",
            "value": "1.00001495e-10",
            "uncertainty": "0.000 000 90 e-10",
        },
        "atomic mass constant": {
            "quantity": "atomic mass constant",
            "unit": "kg",
            "value": "1.66053906660e-27",
            "uncertainty": "0.000 000 000 50 e-27",
        },
        "atomic mass constant energy equivalent": {
            "quantity": "atomic mass constant energy equivalent",
            "unit": "J",
            "value": "1.49241808560e-10",
            "uncertainty": "0.000 000 000 45 e-10",
        },
        "atomic mass constant energy equivalent in mev": {
            "quantity": "atomic mass constant energy equivalent in MeV",
            "unit": "MeV",
            "value": "931.49410242",
            "uncertainty": "0.000 000 28",
        },
        "atomic mass unit-electron volt relationship": {
            "quantity": "atomic mass unit-electron volt relationship",
            "unit": "eV",
            "value": "9.3149410242e8",
            "uncertainty": "0.000 000 0028 e8",
        },
        "atomic mass unit-hartree relationship": {
            "quantity": "atomic mass unit-hartree relationship",
            "unit": "E_h",
            "value": "3.4231776874e7",
            "uncertainty": "0.000 000 0010 e7",
        },
        "atomic mass unit-hertz relationship": {
            "quantity": "atomic mass unit-hertz relationship",
            "unit": "Hz",
            "value": "2.25234271871e23",
            "uncertainty": "0.000 000 000 68 e23",
        },
        "atomic mass unit-inverse meter relationship": {
            "quantity": "atomic mass unit-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "7.5130066104e14",
            "uncertainty": "0.000 000 0023 e14",
        },
        "atomic mass unit-joule relationship": {
            "quantity": "atomic mass unit-joule relationship",
            "unit": "J",
            "value": "1.49241808560e-10",
            "uncertainty": "0.000 000 000 45 e-10",
        },
        "atomic mass unit-kelvin relationship": {
            "quantity": "atomic mass unit-kelvin relationship",
            "unit": "K",
            "value": "1.08095401916e13",
            "uncertainty": "0.000 000 000 33 e13",
        },
        "atomic mass unit-kilogram relationship": {
            "quantity": "atomic mass unit-kilogram relationship",
            "unit": "kg",
            "value": "1.66053906660e-27",
            "uncertainty": "0.000 000 000 50 e-27",
        },
        "atomic unit of 1st hyperpolarizability": {
            "quantity": "atomic unit of 1st hyperpolarizability",
            "unit": "C^3 m^3 J^{-2}",
            "value": "3.2063613061e-53",
            "uncertainty": "0.000 000 0015 e-53",
        },
        "atomic unit of 2nd hyperpolarizability": {
            "quantity": "atomic unit of 2nd hyperpolarizability",
            "unit": "C^4 m^4 J^{-3}",
            "value": "6.2353799905e-65",
            "uncertainty": "0.000 000 0038 e-65",
        },
        "atomic unit of action": {
            "quantity": "atomic unit of action",
            "unit": "J s",
            "value": "1.054571817e-34",
            "uncertainty": "(exact)",
        },
        "atomic unit of charge": {
            "quantity": "atomic unit of charge",
            "unit": "C",
            "value": "1.602176634e-19",
            "uncertainty": "(exact)",
        },
        "atomic unit of charge density": {
            "quantity": "atomic unit of charge density",
            "unit": "C m^{-3}",
            "value": "1.08120238457e12",
            "uncertainty": "0.000 000 000 49 e12",
        },
        "atomic unit of current": {
            "quantity": "atomic unit of current",
            "unit": "A",
            "value": "6.623618237510e-3",
            "uncertainty": "0.000 000 000 013 e-3",
        },
        "atomic unit of electric dipole mom.": {
            "quantity": "atomic unit of electric dipole mom.",
            "unit": "C m",
            "value": "8.4783536255e-30",
            "uncertainty": "0.000 000 0013 e-30",
        },
        "atomic unit of electric field": {
            "quantity": "atomic unit of electric field",
            "unit": "V m^{-1}",
            "value": "5.14220674763e11",
            "uncertainty": "0.000 000 000 78 e11",
        },
        "atomic unit of electric field gradient": {
            "quantity": "atomic unit of electric field gradient",
            "unit": "V m^{-2}",
            "value": "9.7173624292e21",
            "uncertainty": "0.000 000 0029 e21",
        },
        "atomic unit of electric polarizability": {
            "quantity": "atomic unit of electric polarizability",
            "unit": "C^2 m^2 J^{-1}",
            "value": "1.64877727436e-41",
            "uncertainty": "0.000 000 000 50 e-41",
        },
        "atomic unit of electric potential": {
            "quantity": "atomic unit of electric potential",
            "unit": "V",
            "value": "27.211386245988",
            "uncertainty": "0.000 000 000 053",
        },
        "atomic unit of electric quadrupole mom.": {
            "quantity": "atomic unit of electric quadrupole mom.",
            "unit": "C m^2",
            "value": "4.4865515246e-40",
            "uncertainty": "0.000 000 0014 e-40",
        },
        "atomic unit of energy": {
            "quantity": "atomic unit of energy",
            "unit": "J",
            "value": "4.3597447222071e-18",
            "uncertainty": "0.000 000 000 0085 e-18",
        },
        "atomic unit of force": {
            "quantity": "atomic unit of force",
            "unit": "N",
            "value": "8.2387234983e-8",
            "uncertainty": "0.000 000 0012 e-8",
        },
        "atomic unit of length": {
            "quantity": "atomic unit of length",
            "unit": "m",
            "value": "5.29177210903e-11",
            "uncertainty": "0.000 000 000 80 e-11",
        },
        "atomic unit of mag. dipole mom.": {
            "quantity": "atomic unit of mag. dipole mom.",
            "unit": "J T^{-1}",
            "value": "1.85480201566e-23",
            "uncertainty": "0.000 000 000 56 e-23",
        },
        "atomic unit of mag. flux density": {
            "quantity": "atomic unit of mag. flux density",
            "unit": "T",
            "value": "2.35051756758e5",
            "uncertainty": "0.000 000 000 71 e5",
        },
        "atomic unit of magnetizability": {
            "quantity": "atomic unit of magnetizability",
            "unit": "J T^{-2}",
            "value": "7.8910366008e-29",
            "uncertainty": "0.000 000 0048 e-29",
        },
        "atomic unit of mass": {
            "quantity": "atomic unit of mass",
            "unit": "kg",
            "value": "9.1093837015e-31",
            "uncertainty": "0.000 000 0028 e-31",
        },
        "atomic unit of momentum": {
            "quantity": "atomic unit of momentum",
            "unit": "kg m s^{-1}",
            "value": "1.99285191410e-24",
            "uncertainty": "0.000 000 000 30 e-24",
        },
        "atomic unit of permittivity": {
            "quantity": "atomic unit of permittivity",
            "unit": "F m^{-1}",
            "value": "1.11265005545e-10",
            "uncertainty": "0.000 000 000 17 e-10",
        },
        "atomic unit of time": {
            "quantity": "atomic unit of time",
            "unit": "s",
            "value": "2.4188843265857e-17",
            "uncertainty": "0.000 000 000 0047 e-17",
        },
        "atomic unit of velocity": {
            "quantity": "atomic unit of velocity",
            "unit": "m s^{-1}",
            "value": "2.18769126364e6",
            "uncertainty": "0.000 000 000 33 e6",
        },
        "avogadro constant": {
            "quantity": "Avogadro constant",
            "unit": "mol^{-1}",
            "value": "6.02214076e23",
            "uncertainty": "(exact)",
        },
        "bohr magneton": {
            "quantity": "Bohr magneton",
            "unit": "J T^{-1}",
            "value": "9.2740100783e-24",
            "uncertainty": "0.000 000 0028 e-24",
        },
        "bohr magneton in ev/t": {
            "quantity": "Bohr magneton in eV/T",
            "unit": "eV T^{-1}",
            "value": "5.7883818060e-5",
            "uncertainty": "0.000 000 0017 e-5",
        },
        "bohr magneton in hz/t": {
            "quantity": "Bohr magneton in Hz/T",
            "unit": "Hz T^{-1}",
            "value": "1.39962449361e10",
            "uncertainty": "0.000 000 000 42 e10",
        },
        "bohr magneton in inverse meter per tesla": {
            "quantity": "Bohr magneton in inverse meter per tesla",
            "unit": "m^{-1} T^{-1}",
            "value": "46.686447783",
            "uncertainty": "0.000 000 014",
        },
        "bohr magneton in k/t": {
            "quantity": "Bohr magneton in K/T",
            "unit": "K T^{-1}",
            "value": "0.67171381563",
            "uncertainty": "0.000 000 000 20",
        },
        "bohr radius": {
            "quantity": "Bohr radius",
            "unit": "m",
            "value": "5.29177210903e-11",
            "uncertainty": "0.000 000 000 80 e-11",
        },
        "boltzmann constant": {
            "quantity": "Boltzmann constant",
            "unit": "J K^{-1}",
            "value": "1.380649e-23",
            "uncertainty": "(exact)",
        },
        "boltzmann constant in ev/k": {
            "quantity": "Boltzmann constant in eV/K",
            "unit": "eV K^{-1}",
            "value": "8.617333262e-5",
            "uncertainty": "(exact)",
        },
        "boltzmann constant in hz/k": {
            "quantity": "Boltzmann constant in Hz/K",
            "unit": "Hz K^{-1}",
            "value": "2.083661912e10",
            "uncertainty": "(exact)",
        },
        "boltzmann constant in inverse meter per kelvin": {
            "quantity": "Boltzmann constant in inverse meter per kelvin",
            "unit": "m^{-1} K^{-1}",
            "value": "69.50348004",
            "uncertainty": "(exact)",
        },
        "characteristic impedance of vacuum": {
            "quantity": "characteristic impedance of vacuum",
            "unit": "ohm",
            "value": "376.730313668",
            "uncertainty": "0.000 000 057",
        },
        "classical electron radius": {
            "quantity": "classical electron radius",
            "unit": "m",
            "value": "2.8179403262e-15",
            "uncertainty": "0.000 000 0013 e-15",
        },
        "compton wavelength": {
            "quantity": "Compton wavelength",
            "unit": "m",
            "value": "2.42631023867e-12",
            "uncertainty": "0.000 000 000 73 e-12",
        },
        "conductance quantum": {
            "quantity": "conductance quantum",
            "unit": "S",
            "value": "7.748091729e-5",
            "uncertainty": "(exact)",
        },
        "conventional value of ampere-90": {
            "quantity": "conventional value of ampere-90",
            "unit": "A",
            "value": "1.00000008887",
            "uncertainty": "(exact)",
        },
        "conventional value of coulomb-90": {
            "quantity": "conventional value of coulomb-90",
            "unit": "C",
            "value": "1.00000008887",
            "uncertainty": "(exact)",
        },
        "conventional value of farad-90": {
            "quantity": "conventional value of farad-90",
            "unit": "F",
            "value": "0.99999998220",
            "uncertainty": "(exact)",
        },
        "conventional value of henry-90": {
            "quantity": "conventional value of henry-90",
            "unit": "H",
            "value": "1.00000001779",
            "uncertainty": "(exact)",
        },
        "conventional value of josephson constant": {
            "quantity": "conventional value of Josephson constant",
            "unit": "Hz V^{-1}",
            "value": "483597.9e9",
            "uncertainty": "(exact)",
        },
        "conventional value of ohm-90": {
            "quantity": "conventional value of ohm-90",
            "unit": "ohm",
            "value": "1.00000001779",
            "uncertainty": "(exact)",
        },
        "conventional value of volt-90": {
            "quantity": "conventional value of volt-90",
            "unit": "V",
            "value": "1.00000010666",
            "uncertainty": "(exact)",
        },
        "conventional value of von klitzing constant": {
            "quantity": "conventional value of von Klitzing constant",
            "unit": "ohm",
            "value": "25812.807",
            "uncertainty": "(exact)",
        },
        "conventional value of watt-90": {
            "quantity": "conventional value of watt-90",
            "unit": "W",
            "value": "1.00000019553",
            "uncertainty": "(exact)",
        },
        "copper x unit": {
            "quantity": "Copper x unit",
            "unit": "m",
            "value": "1.00207697e-13",
            "uncertainty": "0.000 000 28 e-13",
        },
        "deuteron-electron mag. mom. ratio": {
            "quantity": "deuteron-electron mag. mom. ratio",
            "unit": "",
            "value": "-4.664345551e-4",
            "uncertainty": "0.000 000 012 e-4",
        },
        "deuteron-electron mass ratio": {
            "quantity": "deuteron-electron mass ratio",
            "unit": "",
            "value": "3670.48296788",
            "uncertainty": "0.000 000 13",
        },
        "deuteron g factor": {
            "quantity": "deuteron g factor",
            "unit": "",
            "value": "0.8574382338",
            "uncertainty": "0.000 000 0022",
        },
        "deuteron mag. mom.": {
            "quantity": "deuteron mag. mom.",
            "unit": "J T^{-1}",
            "value": "4.330735094e-27",
            "uncertainty": "0.000 000 011 e-27",
        },
        "deuteron mag. mom. to bohr magneton ratio": {
            "quantity": "deuteron mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "4.669754570e-4",
            "uncertainty": "0.000 000 012 e-4",
        },
        "deuteron mag. mom. to nuclear magneton ratio": {
            "quantity": "deuteron mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "0.8574382338",
            "uncertainty": "0.000 000 0022",
        },
        "deuteron mass": {
            "quantity": "deuteron mass",
            "unit": "kg",
            "value": "3.3435837724e-27",
            "uncertainty": "0.000 000 0010 e-27",
        },
        "deuteron mass energy equivalent": {
            "quantity": "deuteron mass energy equivalent",
            "unit": "J",
            "value": "3.00506323102e-10",
            "uncertainty": "0.000 000 000 91 e-10",
        },
        "deuteron mass energy equivalent in mev": {
            "quantity": "deuteron mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "1875.61294257",
            "uncertainty": "0.000 000 57",
        },
        "deuteron mass in u": {
            "quantity": "deuteron mass in u",
            "unit": "u",
            "value": "2.013553212745",
            "uncertainty": "0.000 000 000 040",
        },
        "deuteron molar mass": {
            "quantity": "deuteron molar mass",
            "unit": "kg mol^{-1}",
            "value": "2.01355321205e-3",
            "uncertainty": "0.000 000 000 61 e-3",
        },
        "deuteron-neutron mag. mom. ratio": {
            "quantity": "deuteron-neutron mag. mom. ratio",
            "unit": "",
            "value": "-0.44820653",
            "uncertainty": "0.000 000 11",
        },
        "deuteron-proton mag. mom. ratio": {
            "quantity": "deuteron-proton mag. mom. ratio",
            "unit": "",
            "value": "0.30701220939",
            "uncertainty": "0.000 000 000 79",
        },
        "deuteron-proton mass ratio": {
            "quantity": "deuteron-proton mass ratio",
            "unit": "",
            "value": "1.99900750139",
            "uncertainty": "0.000 000 000 11",
        },
        "deuteron relative atomic mass": {
            "quantity": "deuteron relative atomic mass",
            "unit": "",
            "value": "2.013553212745",
            "uncertainty": "0.000 000 000 040",
        },
        "deuteron rms charge radius": {
            "quantity": "deuteron rms charge radius",
            "unit": "m",
            "value": "2.12799e-15",
            "uncertainty": "0.000 74 e-15",
        },
        "electron charge to mass quotient": {
            "quantity": "electron charge to mass quotient",
            "unit": "C kg^{-1}",
            "value": "-1.75882001076e11",
            "uncertainty": "0.000 000 000 53 e11",
        },
        "electron-deuteron mag. mom. ratio": {
            "quantity": "electron-deuteron mag. mom. ratio",
            "unit": "",
            "value": "-2143.9234915",
            "uncertainty": "0.000 0056",
        },
        "electron-deuteron mass ratio": {
            "quantity": "electron-deuteron mass ratio",
            "unit": "",
            "value": "2.724437107462e-4",
            "uncertainty": "0.000 000 000 096 e-4",
        },
        "electron g factor": {
            "quantity": "electron g factor",
            "unit": "",
            "value": "-2.00231930436256",
            "uncertainty": "0.000 000 000 000 35",
        },
        "electron gyromag. ratio": {
            "quantity": "electron gyromag. ratio",
            "unit": "s^{-1} T^{-1}",
            "value": "1.76085963023e11",
            "uncertainty": "0.000 000 000 53 e11",
        },
        "electron gyromag. ratio in mhz/t": {
            "quantity": "electron gyromag. ratio in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "28024.9514242",
            "uncertainty": "0.000 0085",
        },
        "electron-helion mass ratio": {
            "quantity": "electron-helion mass ratio",
            "unit": "",
            "value": "1.819543074573e-4",
            "uncertainty": "0.000 000 000 079 e-4",
        },
        "electron mag. mom.": {
            "quantity": "electron mag. mom.",
            "unit": "J T^{-1}",
            "value": "-9.2847647043e-24",
            "uncertainty": "0.000 000 0028 e-24",
        },
        "electron mag. mom. anomaly": {
            "quantity": "electron mag. mom. anomaly",
            "unit": "",
            "value": "1.15965218128e-3",
            "uncertainty": "0.000 000 000 18 e-3",
        },
        "electron mag. mom. to bohr magneton ratio": {
            "quantity": "electron mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "-1.00115965218128",
            "uncertainty": "0.000 000 000 000 18",
        },
        "electron mag. mom. to nuclear magneton ratio": {
            "quantity": "electron mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "-1838.28197188",
            "uncertainty": "0.000 000 11",
        },
        "electron mass": {
            "quantity": "electron mass",
            "unit": "kg",
            "value": "9.1093837015e-31",
            "uncertainty": "0.000 000 0028 e-31",
        },
        "electron mass energy equivalent": {
            "quantity": "electron mass energy equivalent",
            "unit": "J",
            "value": "8.1871057769e-14",
            "uncertainty": "0.000 000 0025 e-14",
        },
        "electron mass energy equivalent in mev": {
            "quantity": "electron mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "0.51099895000",
            "uncertainty": "0.000 000 000 15",
        },
        "electron mass in u": {
            "quantity": "electron mass in u",
            "unit": "u",
            "value": "5.48579909065e-4",
            "uncertainty": "0.000 000 000 16 e-4",
        },
        "electron molar mass": {
            "quantity": "electron molar mass",
            "unit": "kg mol^{-1}",
            "value": "5.4857990888e-7",
            "uncertainty": "0.000 000 0017 e-7",
        },
        "electron-muon mag. mom. ratio": {
            "quantity": "electron-muon mag. mom. ratio",
            "unit": "",
            "value": "206.7669883",
            "uncertainty": "0.000 0046",
        },
        "electron-muon mass ratio": {
            "quantity": "electron-muon mass ratio",
            "unit": "",
            "value": "4.83633169e-3",
            "uncertainty": "0.000 000 11 e-3",
        },
        "electron-neutron mag. mom. ratio": {
            "quantity": "electron-neutron mag. mom. ratio",
            "unit": "",
            "value": "960.92050",
            "uncertainty": "0.000 23",
        },
        "electron-neutron mass ratio": {
            "quantity": "electron-neutron mass ratio",
            "unit": "",
            "value": "5.4386734424e-4",
            "uncertainty": "0.000 000 0026 e-4",
        },
        "electron-proton mag. mom. ratio": {
            "quantity": "electron-proton mag. mom. ratio",
            "unit": "",
            "value": "-658.21068789",
            "uncertainty": "0.000 000 20",
        },
        "electron-proton mass ratio": {
            "quantity": "electron-proton mass ratio",
            "unit": "",
            "value": "5.44617021487e-4",
            "uncertainty": "0.000 000 000 33 e-4",
        },
        "electron relative atomic mass": {
            "quantity": "electron relative atomic mass",
            "unit": "",
            "value": "5.48579909065e-4",
            "uncertainty": "0.000 000 000 16 e-4",
        },
        "electron-tau mass ratio": {
            "quantity": "electron-tau mass ratio",
            "unit": "",
            "value": "2.87585e-4",
            "uncertainty": "0.000 19 e-4",
        },
        "electron to alpha particle mass ratio": {
            "quantity": "electron to alpha particle mass ratio",
            "unit": "",
            "value": "1.370933554787e-4",
            "uncertainty": "0.000 000 000 045 e-4",
        },
        "electron to shielded helion mag. mom. ratio": {
            "quantity": "electron to shielded helion mag. mom. ratio",
            "unit": "",
            "value": "864.058257",
            "uncertainty": "0.000 010",
        },
        "electron to shielded proton mag. mom. ratio": {
            "quantity": "electron to shielded proton mag. mom. ratio",
            "unit": "",
            "value": "-658.2275971",
            "uncertainty": "0.000 0072",
        },
        "electron-triton mass ratio": {
            "quantity": "electron-triton mass ratio",
            "unit": "",
            "value": "1.819200062251e-4",
            "uncertainty": "0.000 000 000 090 e-4",
        },
        "electron volt": {
            "quantity": "electron volt",
            "unit": "J",
            "value": "1.602176634e-19",
            "uncertainty": "(exact)",
        },
        "electron volt-atomic mass unit relationship": {
            "quantity": "electron volt-atomic mass unit relationship",
            "unit": "u",
            "value": "1.07354410233e-9",
            "uncertainty": "0.000 000 000 32 e-9",
        },
        "electron volt-hartree relationship": {
            "quantity": "electron volt-hartree relationship",
            "unit": "E_h",
            "value": "3.6749322175655e-2",
            "uncertainty": "0.000 000 000 0071 e-2",
        },
        "electron volt-hertz relationship": {
            "quantity": "electron volt-hertz relationship",
            "unit": "Hz",
            "value": "2.417989242e14",
            "uncertainty": "(exact)",
        },
        "electron volt-inverse meter relationship": {
            "quantity": "electron volt-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "8.065543937e5",
            "uncertainty": "(exact)",
        },
        "electron volt-joule relationship": {
            "quantity": "electron volt-joule relationship",
            "unit": "J",
            "value": "1.602176634e-19",
            "uncertainty": "(exact)",
        },
        "electron volt-kelvin relationship": {
            "quantity": "electron volt-kelvin relationship",
            "unit": "K",
            "value": "1.160451812e4",
            "uncertainty": "(exact)",
        },
        "electron volt-kilogram relationship": {
            "quantity": "electron volt-kilogram relationship",
            "unit": "kg",
            "value": "1.782661921e-36",
            "uncertainty": "(exact)",
        },
        "elementary charge": {
            "quantity": "elementary charge",
            "unit": "C",
            "value": "1.602176634e-19",
            "uncertainty": "(exact)",
        },
        "elementary charge over h-bar": {
            "quantity": "elementary charge over h-bar",
            "unit": "A J^{-1}",
            "value": "1.519267447e15",
            "uncertainty": "(exact)",
        },
        "faraday constant": {
            "quantity": "Faraday constant",
            "unit": "C mol^{-1}",
            "value": "96485.33212",
            "uncertainty": "(exact)",
        },
        "fermi coupling constant": {
            "quantity": "Fermi coupling constant",
            "unit": "GeV^{-2}",
            "value": "1.1663787e-5",
            "uncertainty": "0.000 0006 e-5",
        },
        "fine-structure constant": {
            "quantity": "fine-structure constant",
            "unit": "",
            "value": "7.2973525693e-3",
            "uncertainty": "0.000 000 0011 e-3",
        },
        "first radiation constant": {
            "quantity": "first radiation constant",
            "unit": "W m^2",
            "value": "3.741771852e-16",
            "uncertainty": "(exact)",
        },
        "first radiation constant for spectral radiance": {
            "quantity": "first radiation constant for spectral radiance",
            "unit": "W m^2 sr^{-1}",
            "value": "1.191042972e-16",
            "uncertainty": "(exact)",
        },
        "hartree-atomic mass unit relationship": {
            "quantity": "hartree-atomic mass unit relationship",
            "unit": "u",
            "value": "2.92126232205e-8",
            "uncertainty": "0.000 000 000 88 e-8",
        },
        "hartree-electron volt relationship": {
            "quantity": "hartree-electron volt relationship",
            "unit": "eV",
            "value": "27.211386245988",
            "uncertainty": "0.000 000 000 053",
        },
        "hartree energy": {
            "quantity": "Hartree energy",
            "unit": "J",
            "value": "4.3597447222071e-18",
            "uncertainty": "0.000 000 000 0085 e-18",
        },
        "hartree energy in ev": {
            "quantity": "Hartree energy in eV",
            "unit": "eV",
            "value": "27.211386245988",
            "uncertainty": "0.000 000 000 053",
        },
        "hartree-hertz relationship": {
            "quantity": "hartree-hertz relationship",
            "unit": "Hz",
            "value": "6.579683920502e15",
            "uncertainty": "0.000 000 000 013 e15",
        },
        "hartree-inverse meter relationship": {
            "quantity": "hartree-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "2.1947463136320e7",
            "uncertainty": "0.000 000 000 0043 e7",
        },
        "hartree-joule relationship": {
            "quantity": "hartree-joule relationship",
            "unit": "J",
            "value": "4.3597447222071e-18",
            "uncertainty": "0.000 000 000 0085 e-18",
        },
        "hartree-kelvin relationship": {
            "quantity": "hartree-kelvin relationship",
            "unit": "K",
            "value": "3.1577502480407e5",
            "uncertainty": "0.000 000 000 0061 e5",
        },
        "hartree-kilogram relationship": {
            "quantity": "hartree-kilogram relationship",
            "unit": "kg",
            "value": "4.8508702095432e-35",
            "uncertainty": "0.000 000 000 0094 e-35",
        },
        "helion-electron mass ratio": {
            "quantity": "helion-electron mass ratio",
            "unit": "",
            "value": "5495.88528007",
            "uncertainty": "0.000 000 24",
        },
        "helion g factor": {
            "quantity": "helion g factor",
            "unit": "",
            "value": "-4.255250615",
            "uncertainty": "0.000 000 050",
        },
        "helion mag. mom.": {
            "quantity": "helion mag. mom.",
            "unit": "J T^{-1}",
            "value": "-1.074617532e-26",
            "uncertainty": "0.000 000 013 e-26",
        },
        "helion mag. mom. to bohr magneton ratio": {
            "quantity": "helion mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "-1.158740958e-3",
            "uncertainty": "0.000 000 014 e-3",
        },
        "helion mag. mom. to nuclear magneton ratio": {
            "quantity": "helion mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "-2.127625307",
            "uncertainty": "0.000 000 025",
        },
        "helion mass": {
            "quantity": "helion mass",
            "unit": "kg",
            "value": "5.0064127796e-27",
            "uncertainty": "0.000 000 0015 e-27",
        },
        "helion mass energy equivalent": {
            "quantity": "helion mass energy equivalent",
            "unit": "J",
            "value": "4.4995394125e-10",
            "uncertainty": "0.000 000 0014 e-10",
        },
        "helion mass energy equivalent in mev": {
            "quantity": "helion mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "2808.39160743",
            "uncertainty": "0.000 000 85",
        },
        "helion mass in u": {
            "quantity": "helion mass in u",
            "unit": "u",
            "value": "3.014932247175",
            "uncertainty": "0.000 000 000 097",
        },
        "helion molar mass": {
            "quantity": "helion molar mass",
            "unit": "kg mol^{-1}",
            "value": "3.01493224613e-3",
            "uncertainty": "0.000 000 000 91 e-3",
        },
        "helion-proton mass ratio": {
            "quantity": "helion-proton mass ratio",
            "unit": "",
            "value": "2.99315267167",
            "uncertainty": "0.000 000 000 13",
        },
        "helion relative atomic mass": {
            "quantity": "helion relative atomic mass",
            "unit": "",
            "value": "3.014932247175",
            "uncertainty": "0.000 000 000 097",
        },
        "helion shielding shift": {
            "quantity": "helion shielding shift",
            "unit": "",
            "value": "5.996743e-5",
            "uncertainty": "0.000 010 e-5",
        },
        "hertz-atomic mass unit relationship": {
            "quantity": "hertz-atomic mass unit relationship",
            "unit": "u",
            "value": "4.4398216652e-24",
            "uncertainty": "0.000 000 0013 e-24",
        },
        "hertz-electron volt relationship": {
            "quantity": "hertz-electron volt relationship",
            "unit": "eV",
            "value": "4.135667696e-15",
            "uncertainty": "(exact)",
        },
        "hertz-hartree relationship": {
            "quantity": "hertz-hartree relationship",
            "unit": "E_h",
            "value": "1.5198298460570e-16",
            "uncertainty": "0.000 000 000 0029 e-16",
        },
        "hertz-inverse meter relationship": {
            "quantity": "hertz-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "3.335640951e-9",
            "uncertainty": "(exact)",
        },
        "hertz-joule relationship": {
            "quantity": "hertz-joule relationship",
            "unit": "J",
            "value": "6.62607015e-34",
            "uncertainty": "(exact)",
        },
        "hertz-kelvin relationship": {
            "quantity": "hertz-kelvin relationship",
            "unit": "K",
            "value": "4.799243073e-11",
            "uncertainty": "(exact)",
        },
        "hertz-kilogram relationship": {
            "quantity": "hertz-kilogram relationship",
            "unit": "kg",
            "value": "7.372497323e-51",
            "uncertainty": "(exact)",
        },
        "hyperfine transition frequency of cs-133": {
            "quantity": "hyperfine transition frequency of Cs-133",
            "unit": "Hz",
            "value": "9192631770",
            "uncertainty": "(exact)",
        },
        "inverse fine-structure constant": {
            "quantity": "inverse fine-structure constant",
            "unit": "",
            "value": "137.035999084",
            "uncertainty": "0.000 000 021",
        },
        "inverse meter-atomic mass unit relationship": {
            "quantity": "inverse meter-atomic mass unit relationship",
            "unit": "u",
            "value": "1.33102505010e-15",
            "uncertainty": "0.000 000 000 40 e-15",
        },
        "inverse meter-electron volt relationship": {
            "quantity": "inverse meter-electron volt relationship",
            "unit": "eV",
            "value": "1.239841984e-6",
            "uncertainty": "(exact)",
        },
        "inverse meter-hartree relationship": {
            "quantity": "inverse meter-hartree relationship",
            "unit": "E_h",
            "value": "4.5563352529120e-8",
            "uncertainty": "0.000 000 000 0088 e-8",
        },
        "inverse meter-hertz relationship": {
            "quantity": "inverse meter-hertz relationship",
            "unit": "Hz",
            "value": "299792458",
            "uncertainty": "(exact)",
        },
        "inverse meter-joule relationship": {
            "quantity": "inverse meter-joule relationship",
            "unit": "J",
            "value": "1.986445857e-25",
            "uncertainty": "(exact)",
        },
        "inverse meter-kelvin relationship": {
            "quantity": "inverse meter-kelvin relationship",
            "unit": "K",
            "value": "1.438776877e-2",
            "uncertainty": "(exact)",
        },
        "inverse meter-kilogram relationship": {
            "quantity": "inverse meter-kilogram relationship",
            "unit": "kg",
            "value": "2.210219094e-42",
            "uncertainty": "(exact)",
        },
        "inverse of conductance quantum": {
            "quantity": "inverse of conductance quantum",
            "unit": "ohm",
            "value": "12906.40372",
            "uncertainty": "(exact)",
        },
        "josephson constant": {
            "quantity": "Josephson constant",
            "unit": "Hz V^{-1}",
            "value": "483597.8484e9",
            "uncertainty": "(exact)",
        },
        "joule-atomic mass unit relationship": {
            "quantity": "joule-atomic mass unit relationship",
            "unit": "u",
            "value": "6.7005352565e9",
            "uncertainty": "0.000 000 0020 e9",
        },
        "joule-electron volt relationship": {
            "quantity": "joule-electron volt relationship",
            "unit": "eV",
            "value": "6.241509074e18",
            "uncertainty": "(exact)",
        },
        "joule-hartree relationship": {
            "quantity": "joule-hartree relationship",
            "unit": "E_h",
            "value": "2.2937122783963e17",
            "uncertainty": "0.000 000 000 0045 e17",
        },
        "joule-hertz relationship": {
            "quantity": "joule-hertz relationship",
            "unit": "Hz",
            "value": "1.509190179e33",
            "uncertainty": "(exact)",
        },
        "joule-inverse meter relationship": {
            "quantity": "joule-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "5.034116567e24",
            "uncertainty": "(exact)",
        },
        "joule-kelvin relationship": {
            "quantity": "joule-kelvin relationship",
            "unit": "K",
            "value": "7.242970516e22",
            "uncertainty": "(exact)",
        },
        "joule-kilogram relationship": {
            "quantity": "joule-kilogram relationship",
            "unit": "kg",
            "value": "1.112650056e-17",
            "uncertainty": "(exact)",
        },
        "kelvin-atomic mass unit relationship": {
            "quantity": "kelvin-atomic mass unit relationship",
            "unit": "u",
            "value": "9.2510873014e-14",
            "uncertainty": "0.000 000 0028 e-14",
        },
        "kelvin-electron volt relationship": {
            "quantity": "kelvin-electron volt relationship",
            "unit": "eV",
            "value": "8.617333262e-5",
            "uncertainty": "(exact)",
        },
        "kelvin-hartree relationship": {
            "quantity": "kelvin-hartree relationship",
            "unit": "E_h",
            "value": "3.1668115634556e-6",
            "uncertainty": "0.000 000 000 0061 e-6",
        },
        "kelvin-hertz relationship": {
            "quantity": "kelvin-hertz relationship",
            "unit": "Hz",
            "value": "2.083661912e10",
            "uncertainty": "(exact)",
        },
        "kelvin-inverse meter relationship": {
            "quantity": "kelvin-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "69.50348004",
            "uncertainty": "(exact)",
        },
        "kelvin-joule relationship": {
            "quantity": "kelvin-joule relationship",
            "unit": "J",
            "value": "1.380649e-23",
            "uncertainty": "(exact)",
        },
        "kelvin-kilogram relationship": {
            "quantity": "kelvin-kilogram relationship",
            "unit": "kg",
            "value": "1.536179187e-40",
            "uncertainty": "(exact)",
        },
        "kilogram-atomic mass unit relationship": {
            "quantity": "kilogram-atomic mass unit relationship",
            "unit": "u",
            "value": "6.0221407621e26",
            "uncertainty": "0.000 000 0018 e26",
        },
        "kilogram-electron volt relationship": {
            "quantity": "kilogram-electron volt relationship",
            "unit": "eV",
            "value": "5.609588603e35",
            "uncertainty": "(exact)",
        },
        "kilogram-hartree relationship": {
            "quantity": "kilogram-hartree relationship",
            "unit": "E_h",
            "value": "2.0614857887409e34",
            "uncertainty": "0.000 000 000 0040 e34",
        },
        "kilogram-hertz relationship": {
            "quantity": "kilogram-hertz relationship",
            "unit": "Hz",
            "value": "1.356392489e50",
            "uncertainty": "(exact)",
        },
        "kilogram-inverse meter relationship": {
            "quantity": "kilogram-inverse meter relationship",
            "unit": "m^{-1}",
            "value": "4.524438335e41",
            "uncertainty": "(exact)",
        },
        "kilogram-joule relationship": {
            "quantity": "kilogram-joule relationship",
            "unit": "J",
            "value": "8.987551787e16",
            "uncertainty": "(exact)",
        },
        "kilogram-kelvin relationship": {
            "quantity": "kilogram-kelvin relationship",
            "unit": "K",
            "value": "6.509657260e39",
            "uncertainty": "(exact)",
        },
        "lattice parameter of silicon": {
            "quantity": "lattice parameter of silicon",
            "unit": "m",
            "value": "5.431020511e-10",
            "uncertainty": "0.000 000 089 e-10",
        },
        "lattice spacing of ideal si (220)": {
            "quantity": "lattice spacing of ideal Si (220)",
            "unit": "m",
            "value": "1.920155716e-10",
            "uncertainty": "0.000 000 032 e-10",
        },
        "loschmidt constant (273.15 k, 100 kpa)": {
            "quantity": "Loschmidt constant (273.15 K, 100 kPa)",
            "unit": "m^{-3}",
            "value": "2.651645804e25",
            "uncertainty": "(exact)",
        },
        "loschmidt constant (273.15 k, 101.325 kpa)": {
            "quantity": "Loschmidt constant (273.15 K, 101.325 kPa)",
            "unit": "m^{-3}",
            "value": "2.686780111e25",
            "uncertainty": "(exact)",
        },
        "luminous efficacy": {
            "quantity": "luminous efficacy",
            "unit": "lm W^{-1}",
            "value": "683",
            "uncertainty": "(exact)",
        },
        "mag. flux quantum": {
            "quantity": "mag. flux quantum",
            "unit": "Wb",
            "value": "2.067833848e-15",
            "uncertainty": "(exact)",
        },
        "molar gas constant": {
            "quantity": "molar gas constant",
            "unit": "J mol^{-1} K^{-1}",
            "value": "8.314462618",
            "uncertainty": "(exact)",
        },
        "molar mass constant": {
            "quantity": "molar mass constant",
            "unit": "kg mol^{-1}",
            "value": "0.99999999965e-3",
            "uncertainty": "0.000 000 000 30 e-3",
        },
        "molar mass of carbon-12": {
            "quantity": "molar mass of carbon-12",
            "unit": "kg mol^{-1}",
            "value": "11.9999999958e-3",
            "uncertainty": "0.000 000 0036 e-3",
        },
        "molar planck constant": {
            "quantity": "molar Planck constant",
            "unit": "J Hz^{-1} mol^{-1}",
            "value": "3.990312712e-10",
            "uncertainty": "(exact)",
        },
        "molar volume of ideal gas (273.15 k, 100 kpa)": {
            "quantity": "molar volume of ideal gas (273.15 K, 100 kPa)",
            "unit": "m^3 mol^{-1}",
            "value": "22.71095464e-3",
            "uncertainty": "(exact)",
        },
        "molar volume of ideal gas (273.15 k, 101.325 kpa)": {
            "quantity": "molar volume of ideal gas (273.15 K, 101.325 kPa)",
            "unit": "m^3 mol^{-1}",
            "value": "22.41396954e-3",
            "uncertainty": "(exact)",
        },
        "molar volume of silicon": {
            "quantity": "molar volume of silicon",
            "unit": "m^3 mol^{-1}",
            "value": "1.205883199e-5",
            "uncertainty": "0.000 000 060 e-5",
        },
        "molybdenum x unit": {
            "quantity": "Molybdenum x unit",
            "unit": "m",
            "value": "1.00209952e-13",
            "uncertainty": "0.000 000 53 e-13",
        },
        "muon compton wavelength": {
            "quantity": "muon Compton wavelength",
            "unit": "m",
            "value": "1.173444110e-14",
            "uncertainty": "0.000 000 026 e-14",
        },
        "muon-electron mass ratio": {
            "quantity": "muon-electron mass ratio",
            "unit": "",
            "value": "206.7682830",
            "uncertainty": "0.000 0046",
        },
        "muon g factor": {
            "quantity": "muon g factor",
            "unit": "",
            "value": "-2.0023318418",
            "uncertainty": "0.000 000 0013",
        },
        "muon mag. mom.": {
            "quantity": "muon mag. mom.",
            "unit": "J T^{-1}",
            "value": "-4.49044830e-26",
            "uncertainty": "0.000 000 10 e-26",
        },
        "muon mag. mom. anomaly": {
            "quantity": "muon mag. mom. anomaly",
            "unit": "",
            "value": "1.16592089e-3",
            "uncertainty": "0.000 000 63 e-3",
        },
        "muon mag. mom. to bohr magneton ratio": {
            "quantity": "muon mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "-4.84197047e-3",
            "uncertainty": "0.000 000 11 e-3",
        },
        "muon mag. mom. to nuclear magneton ratio": {
            "quantity": "muon mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "-8.89059703",
            "uncertainty": "0.000 000 20",
        },
        "muon mass": {
            "quantity": "muon mass",
            "unit": "kg",
            "value": "1.883531627e-28",
            "uncertainty": "0.000 000 042 e-28",
        },
        "muon mass energy equivalent": {
            "quantity": "muon mass energy equivalent",
            "unit": "J",
            "value": "1.692833804e-11",
            "uncertainty": "0.000 000 038 e-11",
        },
        "muon mass energy equivalent in mev": {
            "quantity": "muon mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "105.6583755",
            "uncertainty": "0.000 0023",
        },
        "muon mass in u": {
            "quantity": "muon mass in u",
            "unit": "u",
            "value": "0.1134289259",
            "uncertainty": "0.000 000 0025",
        },
        "muon molar mass": {
            "quantity": "muon molar mass",
            "unit": "kg mol^{-1}",
            "value": "1.134289259e-4",
            "uncertainty": "0.000 000 025 e-4",
        },
        "muon-neutron mass ratio": {
            "quantity": "muon-neutron mass ratio",
            "unit": "",
            "value": "0.1124545170",
            "uncertainty": "0.000 000 0025",
        },
        "muon-proton mag. mom. ratio": {
            "quantity": "muon-proton mag. mom. ratio",
            "unit": "",
            "value": "-3.183345142",
            "uncertainty": "0.000 000 071",
        },
        "muon-proton mass ratio": {
            "quantity": "muon-proton mass ratio",
            "unit": "",
            "value": "0.1126095264",
            "uncertainty": "0.000 000 0025",
        },
        "muon-tau mass ratio": {
            "quantity": "muon-tau mass ratio",
            "unit": "",
            "value": "5.94635e-2",
            "uncertainty": "0.000 40 e-2",
        },
        "natural unit of action": {
            "quantity": "natural unit of action",
            "unit": "J s",
            "value": "1.054571817e-34",
            "uncertainty": "(exact)",
        },
        "natural unit of action in ev s": {
            "quantity": "natural unit of action in eV s",
            "unit": "eV s",
            "value": "6.582119569e-16",
            "uncertainty": "(exact)",
        },
        "natural unit of energy": {
            "quantity": "natural unit of energy",
            "unit": "J",
            "value": "8.1871057769e-14",
            "uncertainty": "0.000 000 0025 e-14",
        },
        "natural unit of energy in mev": {
            "quantity": "natural unit of energy in MeV",
            "unit": "MeV",
            "value": "0.51099895000",
            "uncertainty": "0.000 000 000 15",
        },
        "natural unit of length": {
            "quantity": "natural unit of length",
            "unit": "m",
            "value": "3.8615926796e-13",
            "uncertainty": "0.000 000 0012 e-13",
        },
        "natural unit of mass": {
            "quantity": "natural unit of mass",
            "unit": "kg",
            "value": "9.1093837015e-31",
            "uncertainty": "0.000 000 0028 e-31",
        },
        "natural unit of momentum": {
            "quantity": "natural unit of momentum",
            "unit": "kg m s^{-1}",
            "value": "2.73092453075e-22",
            "uncertainty": "0.000 000 000 82 e-22",
        },
        "natural unit of momentum in mev/c": {
            "quantity": "natural unit of momentum in MeV/c",
            "unit": "MeV/c",
            "value": "0.51099895000",
            "uncertainty": "0.000 000 000 15",
        },
        "natural unit of time": {
            "quantity": "natural unit of time",
            "unit": "s",
            "value": "1.28808866819e-21",
            "uncertainty": "0.000 000 000 39 e-21",
        },
        "natural unit of velocity": {
            "quantity": "natural unit of velocity",
            "unit": "m s^{-1}",
            "value": "299792458",
            "uncertainty": "(exact)",
        },
        "neutron compton wavelength": {
            "quantity": "neutron Compton wavelength",
            "unit": "m",
            "value": "1.31959090581e-15",
            "uncertainty": "0.000 000 000 75 e-15",
        },
        "neutron-electron mag. mom. ratio": {
            "quantity": "neutron-electron mag. mom. ratio",
            "unit": "",
            "value": "1.04066882e-3",
            "uncertainty": "0.000 000 25 e-3",
        },
        "neutron-electron mass ratio": {
            "quantity": "neutron-electron mass ratio",
            "unit": "",
            "value": "1838.68366173",
            "uncertainty": "0.000 000 89",
        },
        "neutron g factor": {
            "quantity": "neutron g factor",
            "unit": "",
            "value": "-3.82608545",
            "uncertainty": "0.000 000 90",
        },
        "neutron gyromag. ratio": {
            "quantity": "neutron gyromag. ratio",
            "unit": "s^{-1} T^{-1}",
            "value": "1.83247171e8",
            "uncertainty": "0.000 000 43 e8",
        },
        "neutron gyromag. ratio in mhz/t": {
            "quantity": "neutron gyromag. ratio in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "29.1646931",
            "uncertainty": "0.000 0069",
        },
        "neutron mag. mom.": {
            "quantity": "neutron mag. mom.",
            "unit": "J T^{-1}",
            "value": "-9.6623651e-27",
            "uncertainty": "0.000 0023 e-27",
        },
        "neutron mag. mom. to bohr magneton ratio": {
            "quantity": "neutron mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "-1.04187563e-3",
            "uncertainty": "0.000 000 25 e-3",
        },
        "neutron mag. mom. to nuclear magneton ratio": {
            "quantity": "neutron mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "-1.91304273",
            "uncertainty": "0.000 000 45",
        },
        "neutron mass": {
            "quantity": "neutron mass",
            "unit": "kg",
            "value": "1.67492749804e-27",
            "uncertainty": "0.000 000 000 95 e-27",
        },
        "neutron mass energy equivalent": {
            "quantity": "neutron mass energy equivalent",
            "unit": "J",
            "value": "1.50534976287e-10",
            "uncertainty": "0.000 000 000 86 e-10",
        },
        "neutron mass energy equivalent in mev": {
            "quantity": "neutron mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "939.56542052",
            "uncertainty": "0.000 000 54",
        },
        "neutron mass in u": {
            "quantity": "neutron mass in u",
            "unit": "u",
            "value": "1.00866491595",
            "uncertainty": "0.000 000 000 49",
        },
        "neutron molar mass": {
            "quantity": "neutron molar mass",
            "unit": "kg mol^{-1}",
            "value": "1.00866491560e-3",
            "uncertainty": "0.000 000 000 57 e-3",
        },
        "neutron-muon mass ratio": {
            "quantity": "neutron-muon mass ratio",
            "unit": "",
            "value": "8.89248406",
            "uncertainty": "0.000 000 20",
        },
        "neutron-proton mag. mom. ratio": {
            "quantity": "neutron-proton mag. mom. ratio",
            "unit": "",
            "value": "-0.68497934",
            "uncertainty": "0.000 000 16",
        },
        "neutron-proton mass difference": {
            "quantity": "neutron-proton mass difference",
            "unit": "kg",
            "value": "2.30557435e-30",
            "uncertainty": "0.000 000 82 e-30",
        },
        "neutron-proton mass difference energy equivalent": {
            "quantity": "neutron-proton mass difference energy equivalent",
            "unit": "J",
            "value": "2.07214689e-13",
            "uncertainty": "0.000 000 74 e-13",
        },
        "neutron-proton mass difference energy equivalent in mev": {
            "quantity": "neutron-proton mass difference energy equivalent in MeV",
            "unit": "MeV",
            "value": "1.29333236",
            "uncertainty": "0.000 000 46",
        },
        "neutron-proton mass difference in u": {
            "quantity": "neutron-proton mass difference in u",
            "unit": "u",
            "value": "1.38844933e-3",
            "uncertainty": "0.000 000 49 e-3",
        },
        "neutron-proton mass ratio": {
            "quantity": "neutron-proton mass ratio",
            "unit": "",
            "value": "1.00137841931",
            "uncertainty": "0.000 000 000 49",
        },
        "neutron relative atomic mass": {
            "quantity": "neutron relative atomic mass",
            "unit": "",
            "value": "1.00866491595",
            "uncertainty": "0.000 000 000 49",
        },
        "neutron-tau mass ratio": {
            "quantity": "neutron-tau mass ratio",
            "unit": "",
            "value": "0.528779",
            "uncertainty": "0.000 036",
        },
        "neutron to shielded proton mag. mom. ratio": {
            "quantity": "neutron to shielded proton mag. mom. ratio",
            "unit": "",
            "value": "-0.68499694",
            "uncertainty": "0.000 000 16",
        },
        "newtonian constant of gravitation": {
            "quantity": "Newtonian constant of gravitation",
            "unit": "m^3 kg^{-1} s^{-2}",
            "value": "6.67430e-11",
            "uncertainty": "0.000 15 e-11",
        },
        "newtonian constant of gravitation over h-bar c": {
            "quantity": "Newtonian constant of gravitation over h-bar c",
            "unit": "(GeV/c^2)^{-2}",
            "value": "6.70883e-39",
            "uncertainty": "0.000 15 e-39",
        },
        "nuclear magneton": {
            "quantity": "nuclear magneton",
            "unit": "J T^{-1}",
            "value": "5.0507837461e-27",
            "uncertainty": "0.000 000 0015 e-27",
        },
        "nuclear magneton in ev/t": {
            "quantity": "nuclear magneton in eV/T",
            "unit": "eV T^{-1}",
            "value": "3.15245125844e-8",
            "uncertainty": "0.000 000 000 96 e-8",
        },
        "nuclear magneton in inverse meter per tesla": {
            "quantity": "nuclear magneton in inverse meter per tesla",
            "unit": "m^{-1} T^{-1}",
            "value": "2.54262341353e-2",
            "uncertainty": "0.000 000 000 78 e-2",
        },
        "nuclear magneton in k/t": {
            "quantity": "nuclear magneton in K/T",
            "unit": "K T^{-1}",
            "value": "3.6582677756e-4",
            "uncertainty": "0.000 000 0011 e-4",
        },
        "nuclear magneton in mhz/t": {
            "quantity": "nuclear magneton in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "7.6225932291",
            "uncertainty": "0.000 000 0023",
        },
        "planck constant": {
            "quantity": "Planck constant",
            "unit": "J Hz^{-1}",
            "value": "6.62607015e-34",
            "uncertainty": "(exact)",
        },
        "planck constant in ev/hz": {
            "quantity": "Planck constant in eV/Hz",
            "unit": "eV Hz^{-1}",
            "value": "4.135667696e-15",
            "uncertainty": "(exact)",
        },
        "planck length": {
            "quantity": "Planck length",
            "unit": "m",
            "value": "1.616255e-35",
            "uncertainty": "0.000 018 e-35",
        },
        "planck mass": {
            "quantity": "Planck mass",
            "unit": "kg",
            "value": "2.176434e-8",
            "uncertainty": "0.000 024 e-8",
        },
        "planck mass energy equivalent in gev": {
            "quantity": "Planck mass energy equivalent in GeV",
            "unit": "GeV",
            "value": "1.220890e19",
            "uncertainty": "0.000 014 e19",
        },
        "planck temperature": {
            "quantity": "Planck temperature",
            "unit": "K",
            "value": "1.416784e32",
            "uncertainty": "0.000 016 e32",
        },
        "planck time": {
            "quantity": "Planck time",
            "unit": "s",
            "value": "5.391247e-44",
            "uncertainty": "0.000 060 e-44",
        },
        "proton charge to mass quotient": {
            "quantity": "proton charge to mass quotient",
            "unit": "C kg^{-1}",
            "value": "9.5788331560e7",
            "uncertainty": "0.000 000 0029 e7",
        },
        "proton compton wavelength": {
            "quantity": "proton Compton wavelength",
            "unit": "m",
            "value": "1.32140985539e-15",
            "uncertainty": "0.000 000 000 40 e-15",
        },
        "proton-electron mass ratio": {
            "quantity": "proton-electron mass ratio",
            "unit": "",
            "value": "1836.15267343",
            "uncertainty": "0.000 000 11",
        },
        "proton g factor": {
            "quantity": "proton g factor",
            "unit": "",
            "value": "5.5856946893",
            "uncertainty": "0.000 000 0016",
        },
        "proton gyromag. ratio": {
            "quantity": "proton gyromag. ratio",
            "unit": "s^{-1} T^{-1}",
            "value": "2.6752218744e8",
            "uncertainty": "0.000 000 0011 e8",
        },
        "proton gyromag. ratio in mhz/t": {
            "quantity": "proton gyromag. ratio in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "42.577478518",
            "uncertainty": "0.000 000 018",
        },
        "proton mag. mom.": {
            "quantity": "proton mag. mom.",
            "unit": "J T^{-1}",
            "value": "1.41060679736e-26",
            "uncertainty": "0.000 000 000 60 e-26",
        },
        "proton mag. mom. to bohr magneton ratio": {
            "quantity": "proton mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "1.52103220230e-3",
            "uncertainty": "0.000 000 000 46 e-3",
        },
        "proton mag. mom. to nuclear magneton ratio": {
            "quantity": "proton mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "2.79284734463",
            "uncertainty": "0.000 000 000 82",
        },
        "proton mag. shielding correction": {
            "quantity": "proton mag. shielding correction",
            "unit": "",
            "value": "2.5689e-5",
            "uncertainty": "0.0011 e-5",
        },
        "proton mass": {
            "quantity": "proton mass",
            "unit": "kg",
            "value": "1.67262192369e-27",
            "uncertainty": "0.000 000 000 51 e-27",
        },
        "proton mass energy equivalent": {
            "quantity": "proton mass energy equivalent",
            "unit": "J",
            "value": "1.50327761598e-10",
            "uncertainty": "0.000 000 000 46 e-10",
        },
        "proton mass energy equivalent in mev": {
            "quantity": "proton mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "938.27208816",
            "uncertainty": "0.000 000 29",
        },
        "proton mass in u": {
            "quantity": "proton mass in u",
            "unit": "u",
            "value": "1.007276466621",
            "uncertainty": "0.000 000 000 053",
        },
        "proton molar mass": {
            "quantity": "proton molar mass",
            "unit": "kg mol^{-1}",
            "value": "1.00727646627e-3",
            "uncertainty": "0.000 000 000 31 e-3",
        },
        "proton-muon mass ratio": {
            "quantity": "proton-muon mass ratio",
            "unit": "",
            "value": "8.88024337",
            "uncertainty": "0.000 000 20",
        },
        "proton-neutron mag. mom. ratio": {
            "quantity": "proton-neutron mag. mom. ratio",
            "unit": "",
            "value": "-1.45989805",
            "uncertainty": "0.000 000 34",
        },
        "proton-neutron mass ratio": {
            "quantity": "proton-neutron mass ratio",
            "unit": "",
            "value": "0.99862347812",
            "uncertainty": "0.000 000 000 49",
        },
        "proton relative atomic mass": {
            "quantity": "proton relative atomic mass",
            "unit": "",
            "value": "1.007276466621",
            "uncertainty": "0.000 000 000 053",
        },
        "proton rms charge radius": {
            "quantity": "proton rms charge radius",
            "unit": "m",
            "value": "8.414e-16",
            "uncertainty": "0.019 e-16",
        },
        "proton-tau mass ratio": {
            "quantity": "proton-tau mass ratio",
            "unit": "",
            "value": "0.528051",
            "uncertainty": "0.000 036",
        },
        "quantum of circulation": {
            "quantity": "quantum of circulation",
            "unit": "m^2 s^{-1}",
            "value": "3.6369475516e-4",
            "uncertainty": "0.000 000 0011 e-4",
        },
        "quantum of circulation times 2": {
            "quantity": "quantum of circulation times 2",
            "unit": "m^2 s^{-1}",
            "value": "7.2738951032e-4",
            "uncertainty": "0.000 000 0022 e-4",
        },
        "reduced compton wavelength": {
            "quantity": "reduced Compton wavelength",
            "unit": "m",
            "value": "3.8615926796e-13",
            "uncertainty": "0.000 000 0012 e-13",
        },
        "reduced muon compton wavelength": {
            "quantity": "reduced muon Compton wavelength",
            "unit": "m",
            "value": "1.867594306e-15",
            "uncertainty": "0.000 000 042 e-15",
        },
        "reduced neutron compton wavelength": {
            "quantity": "reduced neutron Compton wavelength",
            "unit": "m",
            "value": "2.1001941552e-16",
            "uncertainty": "0.000 000 0012 e-16",
        },
        "reduced planck constant": {
            "quantity": "reduced Planck constant",
            "unit": "J s",
            "value": "1.054571817e-34",
            "uncertainty": "(exact)",
        },
        "reduced planck constant in ev s": {
            "quantity": "reduced Planck constant in eV s",
            "unit": "eV s",
            "value": "6.582119569e-16",
            "uncertainty": "(exact)",
        },
        "reduced planck constant times c in mev fm": {
            "quantity": "reduced Planck constant times c in MeV fm",
            "unit": "MeV fm",
            "value": "197.3269804",
            "uncertainty": "(exact)",
        },
        "reduced proton compton wavelength": {
            "quantity": "reduced proton Compton wavelength",
            "unit": "m",
            "value": "2.10308910336e-16",
            "uncertainty": "0.000 000 000 64 e-16",
        },
        "reduced tau compton wavelength": {
            "quantity": "reduced tau Compton wavelength",
            "unit": "m",
            "value": "1.110538e-16",
            "uncertainty": "0.000 075 e-16",
        },
        "rydberg constant": {
            "quantity": "Rydberg constant",
            "unit": "m^{-1}",
            "value": "10973731.568160",
            "uncertainty": "0.000 021",
        },
        "rydberg constant times c in hz": {
            "quantity": "Rydberg constant times c in Hz",
            "unit": "Hz",
            "value": "3.2898419602508e15",
            "uncertainty": "0.000 000 000 0064 e15",
        },
        "rydberg constant times hc in ev": {
            "quantity": "Rydberg constant times hc in eV",
            "unit": "eV",
            "value": "13.605693122994",
            "uncertainty": "0.000 000 000 026",
        },
        "rydberg constant times hc in j": {
            "quantity": "Rydberg constant times hc in J",
            "unit": "J",
            "value": "2.1798723611035e-18",
            "uncertainty": "0.000 000 000 0042 e-18",
        },
        "sackur-tetrode constant (1 k, 100 kpa)": {
            "quantity": "Sackur-Tetrode constant (1 K, 100 kPa)",
            "unit": "",
            "value": "-1.15170753706",
            "uncertainty": "0.000 000 000 45",
        },
        "sackur-tetrode constant (1 k, 101.325 kpa)": {
            "quantity": "Sackur-Tetrode constant (1 K, 101.325 kPa)",
            "unit": "",
            "value": "-1.16487052358",
            "uncertainty": "0.000 000 000 45",
        },
        "second radiation constant": {
            "quantity": "second radiation constant",
            "unit": "m K",
            "value": "1.438776877e-2",
            "uncertainty": "(exact)",
        },
        "shielded helion gyromag. ratio": {
            "quantity": "shielded helion gyromag. ratio",
            "unit": "s^{-1} T^{-1}",
            "value": "2.037894569e8",
            "uncertainty": "0.000 000 024 e8",
        },
        "shielded helion gyromag. ratio in mhz/t": {
            "quantity": "shielded helion gyromag. ratio in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "32.43409942",
            "uncertainty": "0.000 000 38",
        },
        "shielded helion mag. mom.": {
            "quantity": "shielded helion mag. mom.",
            "unit": "J T^{-1}",
            "value": "-1.074553090e-26",
            "uncertainty": "0.000 000 013 e-26",
        },
        "shielded helion mag. mom. to bohr magneton ratio": {
            "quantity": "shielded helion mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "-1.158671471e-3",
            "uncertainty": "0.000 000 014 e-3",
        },
        "shielded helion mag. mom. to nuclear magneton ratio": {
            "quantity": "shielded helion mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "-2.127497719",
            "uncertainty": "0.000 000 025",
        },
        "shielded helion to proton mag. mom. ratio": {
            "quantity": "shielded helion to proton mag. mom. ratio",
            "unit": "",
            "value": "-0.7617665618",
            "uncertainty": "0.000 000 0089",
        },
        "shielded helion to shielded proton mag. mom. ratio": {
            "quantity": "shielded helion to shielded proton mag. mom. ratio",
            "unit": "",
            "value": "-0.7617861313",
            "uncertainty": "0.000 000 0033",
        },
        "shielded proton gyromag. ratio": {
            "quantity": "shielded proton gyromag. ratio",
            "unit": "s^{-1} T^{-1}",
            "value": "2.675153151e8",
            "uncertainty": "0.000 000 029 e8",
        },
        "shielded proton gyromag. ratio in mhz/t": {
            "quantity": "shielded proton gyromag. ratio in MHz/T",
            "unit": "MHz T^{-1}",
            "value": "42.57638474",
            "uncertainty": "0.000 000 46",
        },
        "shielded proton mag. mom.": {
            "quantity": "shielded proton mag. mom.",
            "unit": "J T^{-1}",
            "value": "1.410570560e-26",
            "uncertainty": "0.000 000 015 e-26",
        },
        "shielded proton mag. mom. to bohr magneton ratio": {
            "quantity": "shielded proton mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "1.520993128e-3",
            "uncertainty": "0.000 000 017 e-3",
        },
        "shielded proton mag. mom. to nuclear magneton ratio": {
            "quantity": "shielded proton mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "2.792775599",
            "uncertainty": "0.000 000 030",
        },
        "shielding difference of d and p in hd": {
            "quantity": "shielding difference of d and p in HD",
            "unit": "",
            "value": "2.0200e-8",
            "uncertainty": "0.0020 e-8",
        },
        "shielding difference of t and p in ht": {
            "quantity": "shielding difference of t and p in HT",
            "unit": "",
            "value": "2.4140e-8",
            "uncertainty": "0.0020 e-8",
        },
        "speed of light in vacuum": {
            "quantity": "speed of light in vacuum",
            "unit": "m s^{-1}",
            "value": "299792458",
            "uncertainty": "(exact)",
        },
        "standard acceleration of gravity": {
            "quantity": "standard acceleration of gravity",
            "unit": "m s^{-2}",
            "value": "9.80665",
            "uncertainty": "(exact)",
        },
        "standard atmosphere": {
            "quantity": "standard atmosphere",
            "unit": "Pa",
            "value": "101325",
            "uncertainty": "(exact)",
        },
        "standard-state pressure": {
            "quantity": "standard-state pressure",
            "unit": "Pa",
            "value": "100000",
            "uncertainty": "(exact)",
        },
        "stefan-boltzmann constant": {
            "quantity": "Stefan-Boltzmann constant",
            "unit": "W m^{-2} K^{-4}",
            "value": "5.670374419e-8",
            "uncertainty": "(exact)",
        },
        "tau compton wavelength": {
            "quantity": "tau Compton wavelength",
            "unit": "m",
            "value": "6.97771e-16",
            "uncertainty": "0.000 47 e-16",
        },
        "tau-electron mass ratio": {
            "quantity": "tau-electron mass ratio",
            "unit": "",
            "value": "3477.23",
            "uncertainty": "0.23",
        },
        "tau energy equivalent": {
            "quantity": "tau energy equivalent",
            "unit": "MeV",
            "value": "1776.86",
            "uncertainty": "0.12",
        },
        "tau mass": {
            "quantity": "tau mass",
            "unit": "kg",
            "value": "3.16754e-27",
            "uncertainty": "0.000 21 e-27",
        },
        "tau mass energy equivalent": {
            "quantity": "tau mass energy equivalent",
            "unit": "J",
            "value": "2.84684e-10",
            "uncertainty": "0.000 19 e-10",
        },
        "tau mass in u": {
            "quantity": "tau mass in u",
            "unit": "u",
            "value": "1.90754",
            "uncertainty": "0.000 13",
        },
        "tau molar mass": {
            "quantity": "tau molar mass",
            "unit": "kg mol^{-1}",
            "value": "1.90754e-3",
            "uncertainty": "0.000 13 e-3",
        },
        "tau-muon mass ratio": {
            "quantity": "tau-muon mass ratio",
            "unit": "",
            "value": "16.8170",
            "uncertainty": "0.0011",
        },
        "tau-neutron mass ratio": {
            "quantity": "tau-neutron mass ratio",
            "unit": "",
            "value": "1.89115",
            "uncertainty": "0.000 13",
        },
        "tau-proton mass ratio": {
            "quantity": "tau-proton mass ratio",
            "unit": "",
            "value": "1.89376",
            "uncertainty": "0.000 13",
        },
        "thomson cross section": {
            "quantity": "Thomson cross section",
            "unit": "m^2",
            "value": "6.6524587321e-29",
            "uncertainty": "0.000 000 0060 e-29",
        },
        "triton-electron mass ratio": {
            "quantity": "triton-electron mass ratio",
            "unit": "",
            "value": "5496.92153573",
            "uncertainty": "0.000 000 27",
        },
        "triton g factor": {
            "quantity": "triton g factor",
            "unit": "",
            "value": "5.957924931",
            "uncertainty": "0.000 000 012",
        },
        "triton mag. mom.": {
            "quantity": "triton mag. mom.",
            "unit": "J T^{-1}",
            "value": "1.5046095202e-26",
            "uncertainty": "0.000 000 0030 e-26",
        },
        "triton mag. mom. to bohr magneton ratio": {
            "quantity": "triton mag. mom. to Bohr magneton ratio",
            "unit": "",
            "value": "1.6223936651e-3",
            "uncertainty": "0.000 000 0032 e-3",
        },
        "triton mag. mom. to nuclear magneton ratio": {
            "quantity": "triton mag. mom. to nuclear magneton ratio",
            "unit": "",
            "value": "2.9789624656",
            "uncertainty": "0.000 000 0059",
        },
        "triton mass": {
            "quantity": "triton mass",
            "unit": "kg",
            "value": "5.0073567446e-27",
            "uncertainty": "0.000 000 0015 e-27",
        },
        "triton mass energy equivalent": {
            "quantity": "triton mass energy equivalent",
            "unit": "J",
            "value": "4.5003878060e-10",
            "uncertainty": "0.000 000 0014 e-10",
        },
        "triton mass energy equivalent in mev": {
            "quantity": "triton mass energy equivalent in MeV",
            "unit": "MeV",
            "value": "2808.92113298",
            "uncertainty": "0.000 000 85",
        },
        "triton mass in u": {
            "quantity": "triton mass in u",
            "unit": "u",
            "value": "3.01550071621",
            "uncertainty": "0.000 000 000 12",
        },
        "triton molar mass": {
            "quantity": "triton molar mass",
            "unit": "kg mol^{-1}",
            "value": "3.01550071517e-3",
            "uncertainty": "0.000 000 000 92 e-3",
        },
        "triton-proton mass ratio": {
            "quantity": "triton-proton mass ratio",
            "unit": "",
            "value": "2.99371703414",
            "uncertainty": "0.000 000 000 15",
        },
        "triton relative atomic mass": {
            "quantity": "triton relative atomic mass",
            "unit": "",
            "value": "3.01550071621",
            "uncertainty": "0.000 000 000 12",
        },
        "triton to proton mag. mom. ratio": {
            "quantity": "triton to proton mag. mom. ratio",
            "unit": "",
            "value": "1.0666399191",
            "uncertainty": "0.000 000 0021",
        },
        "unified atomic mass unit": {
            "quantity": "unified atomic mass unit",
            "unit": "kg",
            "value": "1.66053906660e-27",
            "uncertainty": "0.000 000 000 50 e-27",
        },
        "vacuum electric permittivity": {
            "quantity": "vacuum electric permittivity",
            "unit": "F m^{-1}",
            "value": "8.8541878128e-12",
            "uncertainty": "0.000 000 0013 e-12",
        },
        "vacuum mag. permeability": {
            "quantity": "vacuum mag. permeability",
            "unit": "N A^{-2}",
            "value": "1.25663706212e-6",
            "uncertainty": "0.000 000 000 19 e-6",
        },
        "von klitzing constant": {
            "quantity": "von Klitzing constant",
            "unit": "ohm",
            "value": "25812.80745",
            "uncertainty": "(exact)",
        },
        "weak mixing angle": {
            "quantity": "weak mixing angle",
            "unit": "",
            "value": "0.22290",
            "uncertainty": "0.000 30",
        },
        "wien frequency displacement law constant": {
            "quantity": "Wien frequency displacement law constant",
            "unit": "Hz K^{-1}",
            "value": "5.878925757e10",
            "uncertainty": "(exact)",
        },
        "wien wavelength displacement law constant": {
            "quantity": "Wien wavelength displacement law constant",
            "unit": "m K",
            "value": "2.897771955e-3",
            "uncertainty": "(exact)",
        },
        "w to z mass ratio": {
            "quantity": "W to Z mass ratio",
            "unit": "",
            "value": "0.88153",
            "uncertainty": "0.000 17",
        },
    },
}
# fmt: on



import pprint
import re
from typing import Dict, Tuple, Union

from ..exceptions import ChoicesError, MoleculeFormatError, ValidationError
from ..util import filter_comments, provenance_stamp
from . import pubchem
from .from_arrays import from_input_arrays
from .regex import CARTXYZ, CHGMULT, ENDL, NUCLEUS, NUMBER, SEP

__all__ = ["from_string"]


def from_string(
    molstr,
    dtype=None,
    *,
    name=None,
    fix_com=None,
    fix_orientation=None,
    fix_symmetry=None,
    return_processed=False,
    enable_qm=True,
    enable_efp=True,
    missing_enabled_return_qm="none",
    missing_enabled_return_efp="none",
    verbose=1,
) -> Union[Dict, Tuple[Dict, Dict]]:
    """Construct a molecule dictionary from any recognized string format.

    Parameters
    ----------
    molstr : str
        Multiline string specification of molecule in a recognized format.
    dtype : {'xyz', 'xyz+', 'psi4', 'psi4+'}, optional
        Molecule format name; see below for details.
    return_processed : bool, optional
        Additionally return intermediate dictionary.
    enable_qm : bool, optional
        Consider quantum mechanical domain in processing the string constants
        into the returned molrec.
    enable_efp: bool, optional
        Consider effective fragment potential domain in processing the string
        contents into the returned molrec. Only relevant if `dtype` supports EFP.
    missing_enabled_return_qm : {'minimal', 'none', 'error'}
        If `enable_qm=True`, what to do if it has no atoms/fragments?
        Respectively, return a fully valid but empty molrec, return empty
        dictionary, or throw error.
    missing_enabled_return_efp : {'minimal', 'none', 'error'}
        If `enable_efp=True`, what to do if it has no atoms/fragments?
        Respectively, return a fully valid but empty molrec, return empty
        dictionary, or throw error.
    name : str, optional
        Override `molstr` information for label for molecule; should
        be valid Python identifier. One of a very limited number of
        fields (three others follow) for trumping `molstr`. Provided
        for convenience, since the alternative would be collect the
        resulting molrec (discarding the Mol if called from class),
        editing it, then remaking the Mol.
    fix_com : bool, optional
        Override `molstr` information for whether translation of `geom`
        is allowed or disallowed.
    fix_orientation : bool, optional
        Override `molstr` information for whether rotation of `geom`
        is allowed or disallowed.
    fix_symmetry : str, optional
        Override `molstr` information for maximal point group symmetry
        which geometry should be treated.

    Returns
    -------
    molrec : dict
        Molecule dictionary spec. See :py:func:`from_arrays`.
    molinit : dict, optional
        Intermediate "molrec"-like dictionary containing `molstr` info after
        parsing by this function but before the validation and defaulting of
        `from_arrays` that returns the proper `molrec`.
        Only provided if `return_processed` is True.

    Raises
    ------
    qcelemental.MoleculeFormatError
        After processing of `molstr`, only an empty string should remain.
        Anything left is a syntax error.

    Notes
    -----
    Several formats are interpretable ::

        xyz - Strict XYZ format
        -----------------------

            String Layout
            -------------
            <number of atoms>
            comment line
            <element_symbol or atomic_number> <x> <y> <z>
            ...
            <element_symbol or atomic_number> <x> <y> <z>

            QM Domain
            ---------
            Specifiable: geom, elem/elez (element identity)
            Inaccessible: mass, real (vs. ghost), elbl (user label), name, units (assumed [A]),
                          input_units_to_au, fix_com/orientation/symmetry, fragmentation,
                          molecular_charge, molecular_multiplicity

            Notes
            -----
            <number of atoms> is pattern-matched but ignored.

        xyz+ - Enhanced XYZ format
        --------------------------

            String Layout
            -------------
            <number of atoms> [<bohr|au|ang>]
            [<molecular_charge> <molecular_multiplicity>] comment line
            <psi4_nucleus_spec> <x> <y> <z>
            ...
            <psi4_nucleus_spec> <x> <y> <z>

            QM Domain
            ---------
            Specifiable: geom, elem/elez (element identity), mass, real (vs. ghost), elbl (user label),
                         units (defaults [A]), molecular_charge, molecular_multiplicity
            Inaccessible: name, input_units_to_au, fix_com/orientation/symmetry, fragmentation

            Notes
            -----
            <number of atoms> is pattern-matched but ignored.

        psi4 - Psi4 molecule {...} format
        ---------------------------------

            QM Domain
            ---------
            Specifiable: geom, elem/elez (element identity), mass, real (vs. ghost), elbl (user label),
                         units (defaults [A]), fix_com/orientation/symmetry, fragment_separators,
                         fragment_charges, fragment_multiplicities, molecular_charge, molecular_multiplicity
            Inaccessible: name, input_units_to_au

                PubChem
                -------
                pubchem : <cid|name|formula> [*]

                A string like the above searches the PubChem database and substitutes the below. Adding the wildcard
                searches for multiple matches and raises ChoicesError with matches for further consideration attached.

                Specifiable: geom, elem/elez (element identity), units (fixed [A]), molecular_charge,
                             molecular_multiplicity (fixed singlet), name

            EFP Domain
            ----------
            Specifiable: units, fix_com/orientation/symmetry, fragment_files, hint_types, geom_hints
            Inaccessible: anything atomic or fragment details -- geom, elem/elez (element identity),
                          mass, real (vs. ghost), elbl (user label), fragment_separators, fragment_charges,
                          fragment_multiplicities, molecular_charge, molecular_multiplicity

        psi4+ - Psi4 non-Cartesian molecule {...} format
        ------------------------------------------------
        Like `dtype=psi4` (although combination with EFP not tested) except
        that instead of pure-Cartesian geometry, allow variables, zmatrix,
        and un-fully-specified geometries. *Not* MolSSI standard, but we're
        not dropping zmatrix yet. Note that in Psi4 internal coordinates
        defined through a zmatrix have no bearing on geometry
        optimization internals or constraints.

    """
    if verbose >= 2:
        print("<<< FROM_STRING\n", molstr, "\n>>>")

    # << 1 >>  str-->str -- discard comments
    molstr = filter_comments(molstr.strip())

    def parse_as_xyz_ish(molstr, strict):
        molinit = {}

        # << 2 >>  str-->dict -- process atoms, units[, chg, mult]
        molstr, processed = _filter_xyz(molstr, strict=strict)
        molinit.update(processed)

        if molstr:
            raise MoleculeFormatError("""Unprocessable Molecule remanents under {}:\n{}""".format(dtype, molstr))

        return molstr, molinit

    def parse_as_psi4_ish(molstr, unsettled):
        molinit = {}

        # Notes
        # * *_filter functions must fill non-overlapping fields
        # * not recc but can add to downstream by appending to str

        # << 2.1 >>  str-->str -- process pubchem into str for downfunction
        molstr, processed = _filter_pubchem(molstr)
        molinit.update(processed)

        # << 2.2 >>  str-->dict -- process units, com, orient, symm
        molstr, processed = _filter_universals(molstr)
        molinit.update(processed)

        # << 2.3 >>  str-->dict -- process efp frags
        molstr, processed = _filter_libefp(molstr)
        molinit.update(processed)

        # << 2.4 >>  str-->dict -- process atoms, chg, mult, frags
        molstr, processed = _filter_mints(molstr, unsettled=unsettled)
        molinit.update(processed)

        if molstr:
            raise MoleculeFormatError("""Unprocessable Molecule remanents under {}:\n{}""".format(dtype, molstr))

        return molstr, molinit

    if dtype == "xyz":
        molstr, molinit = parse_as_xyz_ish(molstr, strict=True)

    elif dtype == "xyz+":
        molstr, molinit = parse_as_xyz_ish(molstr, strict=False)

    elif dtype == "psi4":
        molstr, molinit = parse_as_psi4_ish(molstr, unsettled=False)

    elif dtype == "psi4+":
        molstr, molinit = parse_as_psi4_ish(molstr, unsettled=True)

    elif dtype is None:
        try:
            molstr, molinit = parse_as_psi4_ish(molstr, unsettled=False)
            dtype = "psi4"
        except MoleculeFormatError:
            try:
                molstr, molinit = parse_as_xyz_ish(molstr, strict=True)
                dtype = "xyz"
            except MoleculeFormatError:
                try:
                    molstr, molinit = parse_as_xyz_ish(molstr, strict=False)
                    dtype = "xyz+"
                except MoleculeFormatError:
                    try:
                        molstr, molinit = parse_as_psi4_ish(molstr, unsettled=True)
                        dtype = "psi4+"
                    except MoleculeFormatError:
                        raise MoleculeFormatError(
                            """Unprocessable Molecule remanents under [psi4, xyz, xyz+, psi4+]:\n{}""".format(molstr)
                        )
    else:
        raise KeyError(f"Molecule: dtype of `{dtype}` not recognized.")

    # << 3 >>  args-->dict -- process name, com, orient, symm from arguments
    processed = _filter_kwargs(name, fix_com, fix_orientation, fix_symmetry)
    molinit.update(processed)

    if verbose >= 2:
        print("\nFROM_STRING (", dtype, ") --> FROM_INPUT_ARRAYS <<<")
        pprint.pprint(molinit)
        print(">>>\n")

    # << 4 >>  dict-->molspec
    molrec = from_input_arrays(
        speclabel=True,
        enable_qm=enable_qm,
        enable_efp=enable_efp,
        missing_enabled_return_qm=missing_enabled_return_qm,
        missing_enabled_return_efp=missing_enabled_return_efp,
        **molinit,
    )

    # replace from_arrays stamp with from_string stamp
    if "qm" in molrec and molrec["qm"]:
        molrec["qm"]["provenance"] = provenance_stamp(__name__)
    if "efp" in molrec and molrec["efp"]:
        molrec["efp"]["provenance"] = provenance_stamp(__name__)

    if verbose >= 2:
        print("\nFROM_STRING MOLREC <<<", molrec, ">>>\n")

    if return_processed:
        return molrec, molinit
    else:
        return molrec


# TODO maybe molrec needs a "fix_loose" flag to signal the reciever can symmetrize
#    pubchemerror = re.compile(r'^\s*PubchemError\s*$', re.IGNORECASE)
#    pubcheminput = re.compile(r'^\s*PubchemInput\s*$', re.IGNORECASE)
#        # N.B. Anything starting with PubchemError will be handled correctly by the molecule parser
#        # in libmints, which will just print the rest of the string and exit gracefully.

pubchemre = re.compile(r"\Apubchem" + r"\s*:\s*" + r"(?P<pubsearch>(([\S ]+)))\Z", re.IGNORECASE)


def _filter_pubchem(string):
    """Find any "pubchem:" lines in `string`, make call to the pubchem database
    and return the XYZ results back to `string`.

    Author: @andysim

    """

    def process_pubchem(matchobj):
        pubsearch = matchobj.group("pubsearch")

        # search pubchem for the provided string
        try:
            results = pubchem.get_pubchem_results(pubsearch)
        except Exception as e:
            raise ValidationError(e.message)

        if pubsearch.endswith("*"):
            pubsearch = pubsearch[:-1]
        if len(results) == 1:
            # There's only 1 result - use it
            xyz = results[0].get_molecule_string()
            processed["name"] = "IUPAC {}".format(results[0].name())
            processed["molecular_charge"] = float(results[0].molecular_charge)
            if "Input Error" in xyz:
                raise ValidationError(xyz)
        else:
            # There are multiple results -- print and exit
            # * formerly, this checked for (then used) any exact match, but now (LAB; Sep 2018), disabling that
            #   since user explicitly added '*' char & "best match" (not available formerly) returned w/o '*'
            msg = "\tPubchemError\n"
            msg += "\tMultiple pubchem results were found. Replace\n\n\t\tpubchem:%s\n\n" % (pubsearch)
            msg += "\twith the Chemical ID number or exact name from one of the following and re-run.\n\n"
            msg += "\t Chemical ID     IUPAC Name\n\n"
            ematches = {}
            for result in results:
                msg += "%s" % (result)
                ematches[result.cid] = result.iupac
            raise ChoicesError(msg, ematches)

        # remove PubchemInput first line and assert [A]
        xyz = xyz.replace("PubchemInput", "units ang")
        return xyz

    reconstitute = []
    processed = {}

    for line in string.split("\n"):
        line = re.sub(pubchemre, process_pubchem, line.strip())
        if line:
            reconstitute.append(line)

    return "\n".join(reconstitute), processed


def _filter_kwargs(name, fix_com, fix_orientation, fix_symmetry):
    processed = {}
    if name is not None:
        processed["name"] = name
    if fix_com is not None:
        processed["fix_com"] = fix_com
    if fix_orientation is not None:
        processed["fix_orientation"] = fix_orientation
    if fix_symmetry is not None:
        processed["fix_symmetry"] = fix_symmetry

    return processed


com = re.compile(r"\A(no_com|nocom)\Z", re.IGNORECASE)
orient = re.compile(r"\A(no_reorient|noreorient)\Z", re.IGNORECASE)
bohrang = re.compile(r"\Aunits?[\s=]+((?P<ubohr>(bohr|au|a.u.))|(?P<uang>(ang|angstrom)))\Z", re.IGNORECASE)
symmetry = re.compile(r"\Asymmetry[\s=]+(?P<pg>\w+)\Z", re.IGNORECASE)


def _filter_universals(string):
    """Process multiline `string` for fix_ and unit markers,
    returning a string of unprocessed `string` and a dictionary of
    processed fields.

    fix_com
    fix_orientation
    fix_symmetry
    #input_units_to_au (not settable)
    units

    """

    def process_com(matchobj):
        processed["fix_com"] = True
        return ""

    def process_orient(matchobj):
        processed["fix_orientation"] = True
        return ""

    def process_bohrang(matchobj):
        if matchobj.group("uang"):
            processed["units"] = "Angstrom"
        elif matchobj.group("ubohr"):
            processed["units"] = "Bohr"
        return ""

    def process_symmetry(matchobj):
        processed["fix_symmetry"] = matchobj.group("pg").lower()
        return ""

    reconstitute = []
    processed = {}
    com_found = False
    orient_found = False
    bohrang_found = False
    symmetry_found = False

    for line in string.split("\n"):
        line = line.strip()
        if not com_found:
            line, com_found = re.subn(com, process_com, line)
        if not orient_found:
            line, orient_found = re.subn(orient, process_orient, line)
        if not bohrang_found:
            line, bohrang_found = re.subn(bohrang, process_bohrang, line)
        if not symmetry_found:
            line, symmetry_found = re.subn(symmetry, process_symmetry, line)
        if line:
            reconstitute.append(line)

    return "\n".join(reconstitute), processed


# fmt: off
fragment_marker = re.compile(r'^\s*--\s*$', re.MULTILINE)
efpxyzabc = re.compile(
    r'\A' + r'efp' + SEP + r'(?P<efpfile>(\w+))' + SEP +
    r'(?P<x>' + NUMBER + r')' + SEP + r'(?P<y>' + NUMBER + r')' + SEP + r'(?P<z>' + NUMBER + r')' + SEP +
    r'(?P<a>' + NUMBER + r')' + SEP + r'(?P<b>' + NUMBER + r')' + SEP + r'(?P<c>' + NUMBER + r')' + ENDL + r'\Z',
    re.IGNORECASE | re.VERBOSE)

efppoints = re.compile(
    r'\A' + r'efp' + SEP + r'(?P<efpfile>(\w+))' + ENDL +
    r'[\s,]*' + r'(?P<x1>' + NUMBER + r')' + SEP + r'(?P<y1>' + NUMBER + r')' + SEP + r'(?P<z1>' + NUMBER + r')' + ENDL +
    r'[\s,]*' + r'(?P<x2>' + NUMBER + r')' + SEP + r'(?P<y2>' + NUMBER + r')' + SEP + r'(?P<z2>' + NUMBER + r')' + ENDL +
    r'[\s,]*' + r'(?P<x3>' + NUMBER + r')' + SEP + r'(?P<y3>' + NUMBER + r')' + SEP + r'(?P<z3>' + NUMBER + r')' + ENDL + r'\Z',
    re.IGNORECASE | re.MULTILINE | re.VERBOSE)
# fmt: on


def _filter_libefp(string):
    def process_efpxyzabc(matchobj):
        processed["fragment_files"].append(matchobj.group("efpfile"))
        processed["hint_types"].append("xyzabc")
        processed["geom_hints"].append(
            [
                float(matchobj.group("x")),
                float(matchobj.group("y")),
                float(matchobj.group("z")),
                float(matchobj.group("a")),
                float(matchobj.group("b")),
                float(matchobj.group("c")),
            ]
        )
        return ""

    def process_efppoints(matchobj):
        processed["fragment_files"].append(matchobj.group("efpfile"))
        processed["hint_types"].append("points")
        processed["geom_hints"].append(
            [
                float(matchobj.group("x1")),
                float(matchobj.group("y1")),
                float(matchobj.group("z1")),
                float(matchobj.group("x2")),
                float(matchobj.group("y2")),
                float(matchobj.group("z2")),
                float(matchobj.group("x3")),
                float(matchobj.group("y3")),
                float(matchobj.group("z3")),
            ]
        )
        return ""

    reconstitute = []
    processed = {}
    processed["fragment_files"] = []
    processed["hint_types"] = []
    processed["geom_hints"] = []

    # handle `--`-demarcated blocks
    for frag in re.split(fragment_marker, string):
        frag = re.sub(efpxyzabc, process_efpxyzabc, frag.strip())
        frag = re.sub(efppoints, process_efppoints, frag)
        if frag:
            reconstitute.append(frag)

    return "\n--\n".join(reconstitute), processed


fragment_marker = re.compile(r"^\s*--\s*$", re.MULTILINE)
cgmp = re.compile(r"\A" + CHGMULT + r"\Z", re.VERBOSE)

VAR = r"(-?[a-z][a-z0-9_]*)"  # slight cheat to allow neg in `variable`
NUCLABEL = r"([A-Z]{1,3}((_\w+)|(\d+))?)"
ANCHORTO = r"((\d+)|" + NUCLABEL + r")"
ANCHORVAL = r"(" + NUMBER + r"|" + VAR + ")"

# fmt: off
atom_cartesian = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + SEP + CARTXYZ + r'\Z',
                            re.IGNORECASE | re.VERBOSE)
atom_vcart = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + SEP +
                        r'(?P<Xval>' + ANCHORVAL + r')' + SEP +
                        r'(?P<Yval>' + ANCHORVAL + r')' + SEP +
                        r'(?P<Zval>' + ANCHORVAL + r')' + r'\Z',
                        re.IGNORECASE | re.VERBOSE)
atom_zmat1 = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + r'\Z',
                        re.IGNORECASE | re.VERBOSE)
atom_zmat2 = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + SEP +
                        r'(?P<Ridx>' + ANCHORTO + r')' + SEP + r'(?P<Rval>' + ANCHORVAL + r')' + r'\Z',
                        re.IGNORECASE | re.VERBOSE)
atom_zmat3 = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + SEP +
                        r'(?P<Ridx>' + ANCHORTO + r')' + SEP + r'(?P<Rval>' + ANCHORVAL + r')' + SEP +
                        r'(?P<Aidx>' + ANCHORTO + r')' + SEP + r'(?P<Aval>' + ANCHORVAL + r')' + r'\Z',
                        re.IGNORECASE | re.VERBOSE)
atom_zmat4 = re.compile(r'\A' + r'(?P<nucleus>' + NUCLEUS + r')' + SEP +
                        r'(?P<Ridx>' + ANCHORTO + r')' + SEP + r'(?P<Rval>' + ANCHORVAL + r')' + SEP +
                        r'(?P<Aidx>' + ANCHORTO + r')' + SEP + r'(?P<Aval>' + ANCHORVAL + r')' + SEP +
                        r'(?P<Didx>' + ANCHORTO + r')' + SEP + r'(?P<Dval>' + ANCHORVAL + r')' + r'\Z',
                        re.IGNORECASE | re.VERBOSE)
variable = re.compile(
    r'\A' + r'(?P<varname>' + VAR + r')' + r'\s*=\s*' + r'(?P<varvalue>((tda)|(' + NUMBER + r')))' + r'\Z',
    re.IGNORECASE | re.VERBOSE,
)
# fmt: on


def _filter_mints(string, unsettled=False):
    """Handle extracting fragment, atom, and chg/mult lines from `string`.

    Returns
    -------
    str, dict
        Returns first a subset (plus some fragment separation guidance) of
            `string` containing the unmatched contents. These are generally input
            violations unless handled by a subsequent processing function.
        Returns second a dictionary with processed extractions. Contains (some
            optional) the following keys.

            molecular_charge : float, optional
            molecular_multiplicity : int, optional
            geom
            elbl
            fragment_separators
            fragment_charges
            fragment_multiplicities

    unsettled : bool, optional
        Whether to allow variable entries and zmat structure, accumulating into
        geom_unsettled, rather than pure numerical Cartesian entries,
        accumulating into geom.

    """

    def process_system_cgmp(matchobj):
        """Handles optional special first fragment with sole contents overall chg/mult."""

        processed["molecular_charge"] = float(matchobj.group("chg"))
        processed["molecular_multiplicity"] = int(matchobj.group("mult"))
        return ""

    def filter_fragment(fstring):
        """Handles extraction from everything within a fragment marker "--" of a
        single chg/mult (or None/None) and multiple atom lines.

        """

        def process_fragment_cgmp(matchobj):
            processed["fragment_charges"].append(float(matchobj.group("chg")))
            processed["fragment_multiplicities"].append(int(matchobj.group("mult")))
            return ""

        def process_atom_cartesian(matchobj):
            processed["elbl"].append(matchobj.group("nucleus"))
            processed["geom"].append(float(matchobj.group("x")))
            processed["geom"].append(float(matchobj.group("y")))
            processed["geom"].append(float(matchobj.group("z")))
            return ""

        def process_atom_unsettled(matchobj):
            processed["elbl"].append(matchobj.group("nucleus"))
            geo = []
            if "Xval" in matchobj.groupdict():
                geo.append(matchobj.group("Xval"))
                geo.append(matchobj.group("Yval"))
                geo.append(matchobj.group("Zval"))
            if "Rval" in matchobj.groupdict():
                geo.append(matchobj.group("Ridx"))
                geo.append(matchobj.group("Rval"))
            if "Aval" in matchobj.groupdict():
                geo.append(matchobj.group("Aidx"))
                geo.append(matchobj.group("Aval"))
            if "Dval" in matchobj.groupdict():
                geo.append(matchobj.group("Didx"))
                geo.append(matchobj.group("Dval"))
            processed["geom_unsettled"].append(geo)
            return ""

        def process_variable(matchobj):
            processed["variables"].append((matchobj.group("varname"), matchobj.group("varvalue")))
            return ""

        freconstitute = []
        start_atom = len(processed["elbl"])
        if start_atom > 0:
            processed["fragment_separators"].append(start_atom)

        fcgmp_found = False
        for iln, line in enumerate(fstring.split("\n")):
            line = line.strip()
            if not fcgmp_found:
                line, fcgmp_found = re.subn(cgmp, process_fragment_cgmp, line)
            if unsettled:
                line = re.sub(atom_vcart, process_atom_unsettled, line)
                line = re.sub(atom_zmat1, process_atom_unsettled, line)
                line = re.sub(atom_zmat2, process_atom_unsettled, line)
                line = re.sub(atom_zmat3, process_atom_unsettled, line)
                line = re.sub(atom_zmat4, process_atom_unsettled, line)
                line = re.sub(variable, process_variable, line)
            else:
                line = re.sub(atom_cartesian, process_atom_cartesian, line)
            if line:
                freconstitute.append(line)

        if not fcgmp_found:
            processed["fragment_charges"].append(None)
            processed["fragment_multiplicities"].append(None)

        return "\n".join(freconstitute), processed

    reconstitute = []
    processed = {}
    processed["elbl"] = []
    processed["fragment_separators"] = []
    processed["fragment_charges"] = []
    processed["fragment_multiplicities"] = []
    if unsettled:
        processed["geom_unsettled"] = []
        processed["variables"] = []
    else:
        processed["geom"] = []

    # handle `--`-demarcated blocks
    for ifr, frag in enumerate(re.split(fragment_marker, string)):
        frag = frag.strip()
        if ifr == 0 and cgmp.match(frag):
            frag, ntotch = re.subn(cgmp, process_system_cgmp, frag)
        else:
            frag, processed = filter_fragment(frag)
        if frag:
            reconstitute.append(frag)

    return "\n--\n".join(reconstitute), processed


xyz1strict = re.compile(r"\A" + r"(?P<nat>\d+)" + r"\Z")
SIMPLENUCLEUS = r"""((?P<E>[A-Z]{1,3})|(?P<Z>\d{1,3}))"""
atom_cartesian_strict = re.compile(
    r"\A" + r"(?P<nucleus>" + SIMPLENUCLEUS + r")" + SEP + CARTXYZ + r"\Z", re.IGNORECASE | re.VERBOSE
)

xyz1 = re.compile(r"\A" + r"(?P<nat>\d+)" + r"[\s,]*" + r"((?P<ubohr>(bohr|au))|(?P<uang>ang))?" + r"\Z", re.IGNORECASE)
xyz2 = re.compile(r"\A" + CHGMULT, re.VERBOSE)
atom_cartesian = re.compile(
    r"\A" + r"(?P<nucleus>" + NUCLEUS + r")" + SEP + CARTXYZ + r"\Z", re.IGNORECASE | re.VERBOSE
)


def _filter_xyz(string, strict):
    """Handle extracting atom, units, and chg/mult lines from `string`.

    Parameters
    ----------
    strict : bool
        Whether to enforce a strict XYZ file format or to allow units, chg/mult,
        and add'l atom info.

    Returns
    -------
    str, dict
        Returns first a subset of `string` containing the unmatched contents.
        These are generally input violations unless handled by a subsequent
        processing function.
        Returns second a dictionary with processed extractions. Contains (some
            optional) the following keys.

            molecular_charge : float, optional (`strict=False` only)
            molecular_multiplicity : int, optional (`strict=False` only)
            geom
            elbl
            units : {'Angstrom', 'Bohr'} (`Bohr` `strict=False` only)

    """

    def process_bohrang(matchobj):
        nat = matchobj.group("nat")  # lgtm[py/unused-local-variable]
        if matchobj.group("uang"):
            processed["units"] = "Angstrom"
        elif matchobj.group("ubohr"):
            processed["units"] = "Bohr"
        return ""

    def process_system_cgmp(matchobj):
        processed["molecular_charge"] = float(matchobj.group("chg"))
        processed["molecular_multiplicity"] = int(matchobj.group("mult"))
        return ""

    def process_atom_cartesian(matchobj):
        processed["elbl"].append(matchobj.group("nucleus"))
        processed["geom"].append(float(matchobj.group("x")))
        processed["geom"].append(float(matchobj.group("y")))
        processed["geom"].append(float(matchobj.group("z")))
        return ""

    # nat = 0
    reconstitute = []
    processed = {}
    processed["geom"] = []
    processed["elbl"] = []

    if strict:
        for iln, line in enumerate(string.split("\n")):
            line = line.strip()
            if iln == 0:
                line = re.sub(xyz1strict, "", line)
            elif iln == 1:
                continue
            else:
                line = re.sub(atom_cartesian_strict, process_atom_cartesian, line)
            if line:
                reconstitute.append(line)
    else:
        for iln, line in enumerate(string.split("\n")):
            line = line.strip()
            if iln == 0:
                line = re.sub(xyz1, process_bohrang, line)
            elif iln == 1:
                line = re.sub(xyz2, process_system_cgmp, line)
            else:
                line = re.sub(atom_cartesian, process_atom_cartesian, line)
            if line and iln != 1:
                reconstitute.append(line)

    if "units" not in processed:
        processed["units"] = "Angstrom"

    # if len(processed['geom']) != nat:
    #    raise ValidationError
    processed["geom_hints"] = []  # no EFP

    return "\n".join(reconstitute), processed



from decimal import Decimal

import numpy as np
import pydantic
import pytest

import qcelemental as qcel
from qcelemental.testing import compare_recursive, compare_values


@pytest.fixture
def dataset():
    datums = {
        "decimal": qcel.Datum(
            "a label", "mdyn/angstrom", Decimal("4.4"), comment="force constant", doi="10.1000/182", numeric=False
        ),
        "ndarray": qcel.Datum("an array", "cm^-1", np.arange(4, dtype=np.float) * 4 / 3, comment="freqs"),
        "float": qcel.Datum("a float", "kg", 4.4, doi="10.1000/182"),
        "string": qcel.Datum("ze lbl", "ze unit", "ze data", numeric=False),
        "lststr": qcel.Datum("ze lbl", "ze unit", ["V", "R", None], numeric=False),
    }

    return datums


def test_creation(dataset):
    datum1 = dataset["decimal"]

    assert datum1.label == "a label"
    assert datum1.units == "mdyn/angstrom"
    assert datum1.data == Decimal("4.4")
    assert datum1.numeric is True  # checking that numeric got properly reset from input


def test_creation_nonnum(dataset):
    datum1 = dataset["string"]

    assert datum1.label == "ze lbl"
    assert datum1.units == "ze unit"
    assert datum1.data == "ze data"
    assert datum1.numeric is False


def test_creation_error():
    with pytest.raises(pydantic.ValidationError):
        qcel.Datum("ze lbl", "ze unit", "ze data")

    # assert 'Datum data should be float' in str(e)


@pytest.mark.parametrize(
    "inp,expected",
    [
        (("decimal", None), 4.4),
        (("decimal", "N/m"), 440),
        (("decimal", "hartree/bohr/bohr"), 0.282614141011 if qcel.constants.name == "CODATA2014" else 0.28261413658),
        (("ndarray", "1/m"), np.arange(4, dtype=np.float) * 400 / 3),
    ],
)
def test_units(dataset, inp, expected):
    assert compare_values(expected, dataset[inp[0]].to_units(inp[1]), atol=1.0e-9)


def test_printing(dataset):
    datum1 = dataset["decimal"]
    str1 = """----------------------------------------
             Datum a label
                 Pytest
----------------------------------------
Data:     4.4
Units:    [mdyn/angstrom]
doi:      10.1000/182
Comment:  force constant
Glossary:
----------------------------------------"""

    # Handle some odd spaces in the assert
    str2 = datum1.__str__(label="Pytest")
    assert all(x == y for x, y in zip(str1.split(), str2.split()))


def test_mass_printing_blank():
    pvnone = """
  Variable Map:
  ----------------------------------------------------------------------------
  (none)"""

    assert pvnone == qcel.datum.print_variables({})


def test_mass_printing(dataset):
    str1 = """
  Variable Map:
  ----------------------------------------------------------------------------
  "decimal" =>                    4.4 [mdyn/angstrom]
  "float"   =>         4.400000000000 [kg]
  "lststr"  =>       ['V', 'R', None] [ze unit]
  "ndarray" =>                        [cm^-1]
        [0.         1.33333333 2.66666667 4.        ]
  "string"  =>                ze data [ze unit]
"""

    assert str1 == qcel.datum.print_variables(dataset)


def test_to_dict(dataset):
    listans = [i * 4 / 3 for i in range(4)]
    ans = {"label": "an array", "units": "cm^-1", "data": listans, "comment": "freqs", "numeric": True}

    dicary = dataset["ndarray"].dict()
    assert compare_recursive(ans, dicary, 9)


def test_complex_scalar():
    datum1 = qcel.Datum("complex scalar", "", complex(1, 2))
    ans = {"label": "complex scalar", "units": "", "data": complex(1, 2), "numeric": True}

    assert datum1.label == "complex scalar"
    assert datum1.units == ""
    assert datum1.data.real == 1
    assert datum1.data.imag == 2

    dicary = datum1.dict()
    assert compare_recursive(ans, dicary, 9)


def test_complex_array():
    datum1 = qcel.Datum("complex array", "", np.arange(3, dtype=np.complex_) + 1j)
    ans = {
        "label": "complex array",
        "units": "",
        "data": [complex(0, 1), complex(1, 1), complex(2, 1)],
        "numeric": True,
    }

    dicary = datum1.dict()
    assert compare_recursive(ans, dicary, 9)



"""Queries the PubChem database using a compound name (i.e. 1,3,5-hexatriene)
   to obtain a molecule string that can be passed to Molecule. ::

      results = getPubChemObj("1,3,5-hexatriene")

      Results is an array of results from PubChem matches to your query.
        for entry in results:
           entry["CID"]         => PubChem compound identifer
           entry["IUPAC"]       => IUPAC name for the resulting compound
           entry["PubChemObj"]  => instance of PubChemObj for this compound

           entry["PubChemObj"].get_molecule_string()   => returns a string compatible
                                                        with Psi4's Molecule creation

"""

import json
import re

from ..exceptions import ValidationError
from .regex import DECIMAL


class PubChemObj:
    def __init__(self, cid, mf, iupac, charge):
        self.url = "http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi"
        self.cid = cid
        self.mf = mf
        self.iupac = iupac
        self.molecular_charge = charge
        self.natom = 0
        self.dataSDF = ""

    def __str__(self):
        return "%17d   %s\n" % (self.cid, self.iupac)

    def get_sdf(self):
        """Function to return the SDF (structure-data file) of the PubChem object."""
        from urllib.request import urlopen, Request
        from urllib.parse import quote
        from urllib.error import URLError

        if len(self.dataSDF) == 0:
            url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{}/SDF?record_type=3d".format(
                quote(str(self.cid))
            )
            req = Request(url, headers={"Accept": "chemical/x-mdl-sdfile"})
            try:
                self.dataSDF = urlopen(req).read().decode("utf-8")
            except URLError as e:
                msg = "Unable to open\n\n%s\n\ndue to the error\n\n%s\n\n" % (url, e)
                msg += "It is possible that 3D information does not exist for this molecule in the PubChem database\n"
                print(msg)
                raise ValidationError(msg)
        return self.dataSDF

    def name(self):
        """Function to return the IUPAC name of the PubChem object."""
        return self.iupac

    def get_cartesian(self):
        """Function to return a string of the atom symbol and XYZ
        coordinates of the PubChem object.

        """
        try:
            sdf_text = self.get_sdf()
        except Exception as e:
            raise ValidationError(e.message)

        # Find
        # NA NB                        CONSTANT
        # 14 13  0     0  0  0  0  0  0999 V2000
        m = re.search(r"^\s*(\d+)\s+(?:\d+\s+){8}V2000$", sdf_text, re.MULTILINE)
        self.natom = 0
        if m:
            self.natom = int(m.group(1))

        if self.natom == 0:
            raise ValidationError(
                "PubChem: Cannot find the number of atoms.  3D data doesn't appear\n"
                + "to be available for %s.\n" % self.iupac
            )

        lines = re.split("\n", sdf_text)

        #  3.7320   -0.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
        atom_re = re.compile(
            r"^\s*" + DECIMAL + r"\s+" + DECIMAL + r"\s+" + DECIMAL + r"\s*(\w+)(?:\s+\d+){12}", re.VERBOSE
        )

        molecule_string = "PubchemInput\n"

        atom_count = 0
        for line in lines:

            atom_match = atom_re.match(line)
            if atom_match:
                x = float(atom_match.group(1))
                y = float(atom_match.group(2))
                z = float(atom_match.group(3))
                sym = atom_match.group(4)

                atom_count = atom_count + 1

                molecule_string += "%s %10.6f %10.6f %10.6f\n" % (sym, x, y, z)

                if atom_count == self.natom:
                    break

        return molecule_string

    def get_molecule_string(self):
        """Function to obtain a molecule string through
        get_cartesian() or fail.
        """
        try:
            return self.get_cartesian()
        except Exception as e:
            return e.message


def get_pubchem_results(name):
    """Function to query the PubChem database for molecules matching the
    input string. Builds a PubChemObj object if found.

    """
    from urllib.request import urlopen
    from urllib.parse import quote
    from urllib.error import URLError

    if name.isdigit():
        print("\tSearching PubChem database for CID {}".format(name))
        url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{}/property/IUPACName,MolecularFormula,Charge/JSON".format(
            quote(name)
        )

    else:
        if name.endswith("*"):
            name = name[:-1]
            loose = True
        else:
            loose = False
        print(
            "\tSearching PubChem database for {} ({} returned)".format(
                name, "all matches" if loose else "single best match"
            )
        )
        url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/{}/property/IUPACName,MolecularFormula,Charge/JSON?name_type={}".format(
            quote(name), "word" if loose else "complete"
        )

    try:
        response = urlopen(url)
    except URLError as e:
        # travis urllib.error.HTTPError: HTTP Error 503: Service Unavailable
        raise ValidationError(
            """\tPubchemError\n%s\n\treceived when trying to open\n\t%s\n\tCheck your internet connection, and the above URL, and try again.\n"""
            % (str(e), url)
        ) from e
    data = json.loads(response.read().decode("utf-8"))
    results = []
    for d in data["PropertyTable"]["Properties"]:
        if "IUPACName" not in d:
            continue
        pubobj = PubChemObj(d["CID"], d["IUPACName"], d["IUPACName"], d["Charge"])
        results.append(pubobj)

    print("\tFound {} result(s)".format(len(results)))
    return results


if __name__ == "__main__":
    # * comment "..exceptions" line above
    # * sulfonate below has no 3D structure available
    # * XYZ printing for tropolone* suppressed b/c too many and some have no 3D

    for inp in [
        "1-methoxy-4-[(E)-prop-1-enyl]benzene",
        "4-[bis(4-hydroxyphenyl)methyl]phenol",
        "tropolone",
        "tropolone*",
        "sodium benzenesulfonate",
    ]:  # pragma: no cover
        obj = get_pubchem_results(inp)

        for r in obj:
            print(r, end="")
            if inp != "tropolone*":
                print(r.get_molecule_string())



"""
Units handlers
"""

# Ensure pint exists
from importlib.util import find_spec

from .context import PhysicalConstantsContext, constants

spec = find_spec("pint")
if spec is None:
    raise ModuleNotFoundError(
        """Python module pint not found. Solve by installing it: `conda install pint -c conda-forge` or `pip install pint`"""
    )  # pragma: no cover
del spec, find_spec



import copy
import logging
import pprint
import sys
from typing import Callable, Dict, List, Union

import numpy as np
from pydantic import BaseModel

pp = pprint.PrettyPrinter(width=120)


def _handle_return(passfail: bool, label: str, message: str, return_message: bool, quiet: bool = False):
    """Function to print a '*label*...PASSED' line to log."""

    if not quiet:
        if passfail:
            logging.info(f"    {label:.<53}PASSED")
        else:
            logging.error(f"    {label:.<53}FAILED")
            logging.error(f"    {message:.<53}")

    if return_message:
        return passfail, message
    else:
        return passfail


def tnm() -> str:
    """Returns the name of the calling function, usually name of test case."""

    return sys._getframe().f_back.f_code.co_name


def compare_values(
    expected,
    computed,
    label: str = None,
    *,
    atol: float = 1.0e-6,
    rtol: float = 1.0e-16,
    equal_nan: bool = False,
    equal_phase: bool = False,
    passnone: bool = False,
    quiet: bool = False,
    return_message: bool = False,
    return_handler: Callable = None,
) -> bool:
    """Returns True if two floats or float arrays are element-wise equal within a tolerance.

    Parameters
    ----------
    expected : float or float array-like
        Reference value against which `computed` is compared.
    computed : float or float array-like
        Input value to compare against `expected`.
    atol : float, optional
        Absolute tolerance (see formula below).
    label : str, optional
        Label for passed and error messages. Defaults to calling function name.
    rtol : float, optional
        Relative tolerance (see formula below). By default set to zero so `atol` dominates.
    equal_nan : bool, optional
        Passed to np.isclose. Compare NaN's as equal.
    equal_phase : bool, optional
        Compare computed *or its opposite* as equal.
    passnone : bool, optional
        Return True when both expected and computed are None.
    quiet : bool, optional
        Whether to log the return message.
    return_message : bool, optional
        Whether to return tuple. See below.

    Returns
    -------
    allclose : bool
        Returns True if `expected` and `computed` are equal within tolerance; False otherwise.
    message : str, optional
        When return_message=True, also return passed or error message.

    Other Parameters
    ----------------
    return_handler : function, optional
        Function to control printing, logging, raising, and returning.
        Specialized interception for interfacing testing systems.

    Notes
    -----
    * Akin to np.allclose.
    * For scalar float-comparable types and for arbitrary-dimension, np.ndarray-castable, uniform-type,
      float-comparable types. For mixed types, use :py:func:`compare_recursive`.
    * Sets rtol to zero to match expected Psi4 behaviour, otherwise measured as:

    .. code-block:: python

        absolute(computed - expected) <= (atol + rtol * absolute(expected))

    """
    label = label or sys._getframe().f_back.f_code.co_name
    pass_message = f"\t{label:.<66}PASSED"
    if return_handler is None:
        return_handler = _handle_return

    if passnone:
        if expected is None and computed is None:
            return return_handler(True, label, pass_message, return_message, quiet)

    if np.iscomplexobj(expected):
        dtype = np.complex
    else:
        dtype = np.float

    try:
        xptd, cptd = np.array(expected, dtype=dtype), np.array(computed, dtype=dtype)
    except Exception:
        return return_handler(
            False, label, f"""\t{label}: inputs not cast-able to ndarray of {dtype}.""", return_message, quiet
        )

    if xptd.shape != cptd.shape:
        return return_handler(
            False,
            label,
            f"""\t{label}: computed shape ({cptd.shape}) does not match ({xptd.shape}).""",
            return_message,
            quiet,
        )  # lgtm: [py/syntax-error]

    digits1 = abs(int(np.log10(atol))) + 2
    digits_str = f"to atol={atol}"
    if rtol > 1.0e-12:
        digits_str += f", rtol={rtol}"

    isclose = np.isclose(cptd, xptd, rtol=rtol, atol=atol, equal_nan=equal_nan)
    allclose = bool(np.all(isclose))

    if not allclose and equal_phase and hasattr(cptd, "__neg__"):
        n_isclose = np.isclose(-cptd, xptd, rtol=rtol, atol=atol, equal_nan=equal_nan)
        allclose = bool(np.all(n_isclose))

    if allclose:
        message = pass_message

    else:
        if xptd.shape == ():
            xptd_str = f"{float(xptd):.{digits1}f}"
        else:
            xptd_str = np.array_str(xptd, max_line_width=120, precision=12, suppress_small=True)
            xptd_str = "\n".join("    " + ln for ln in xptd_str.splitlines())

        if cptd.shape == ():
            cptd_str = f"{float(cptd):.{digits1}f}"
        else:
            cptd_str = np.array_str(cptd, max_line_width=120, precision=12, suppress_small=True)
            cptd_str = "\n".join("    " + ln for ln in cptd_str.splitlines())

        diff = cptd - xptd
        if xptd.shape == ():
            diff_str = f"{float(diff):.{digits1}f}"
            message = """\t{}: computed value ({}) does not match ({}) {} by difference ({}).""".format(
                label, cptd_str, xptd_str, digits_str, diff_str
            )
        else:
            diff[isclose] = 0.0
            diff_str = np.array_str(diff, max_line_width=120, precision=12, suppress_small=False)
            diff_str = "\n".join("    " + ln for ln in diff_str.splitlines())
            message = """\t{}: computed value does not match {}.\n  Expected:\n{}\n  Observed:\n{}\n  Difference (passed elements are zeroed):\n{}\n""".format(
                label, digits_str, xptd_str, cptd_str, diff_str
            )

    return return_handler(allclose, label, message, return_message, quiet)


def compare(
    expected,
    computed,
    label: str = None,
    *,
    equal_phase: bool = False,
    quiet: bool = False,
    return_message: bool = False,
    return_handler: Callable = None,
) -> bool:
    """Returns True if two integers, strings, booleans, or integer arrays are element-wise equal.

    Parameters
    ----------
    expected : int, bool, str or int array-like
        Reference value against which `computed` is compared.
    computed : int, bool, str or int array-like
        Input value to compare against `expected`.
    label : str, optional
        Label for passed and error messages. Defaults to calling function name.
    equal_phase : bool, optional
        Compare computed *or its opposite* as equal.

    Returns
    -------
    allclose : bool
        Returns True if `expected` and `computed` are equal; False otherwise.
    message : str, optional
        When return_message=True, also return passed or error message.

    Other Parameters
    ----------------
    return_handler : function, optional
        Function to control printing, logging, raising, and returning.
        Specialized interception for interfacing testing systems.

    Notes
    -----
    * Akin to np.array_equal.
    * For scalar exactly-comparable types and for arbitrary-dimension, np.ndarray-castable, uniform-type,
      exactly-comparable types. For mixed types, use :py:func:`compare_recursive`.

    """
    label = label or sys._getframe().f_back.f_code.co_name
    pass_message = f"\t{label:.<66}PASSED"
    if return_handler is None:
        return_handler = _handle_return

    try:
        xptd, cptd = np.array(expected), np.array(computed)
    except Exception:
        return return_handler(False, label, f"""\t{label}: inputs not cast-able to ndarray.""", return_message, quiet)

    if xptd.shape != cptd.shape:
        return return_handler(
            False,
            label,
            f"""\t{label}: computed shape ({cptd.shape}) does not match ({xptd.shape}).""",
            return_message,
            quiet,
        )

    isclose = np.asarray(xptd == cptd)
    allclose = bool(isclose.all())

    if not allclose and equal_phase:
        try:
            n_isclose = np.asarray(xptd == -cptd)
        except TypeError:
            pass
        else:
            allclose = bool(n_isclose.all())

    if allclose:
        message = pass_message

    else:
        if xptd.shape == ():
            xptd_str = f"{xptd}"
        else:
            xptd_str = np.array_str(xptd, max_line_width=120, precision=12, suppress_small=True)
            xptd_str = "\n".join("    " + ln for ln in xptd_str.splitlines())

        if cptd.shape == ():
            cptd_str = f"{cptd}"
        else:
            cptd_str = np.array_str(cptd, max_line_width=120, precision=12, suppress_small=True)
            cptd_str = "\n".join("    " + ln for ln in cptd_str.splitlines())

        try:
            diff = cptd - xptd
        except TypeError:
            diff_str = "(n/a)"
        else:
            if xptd.shape == ():
                diff_str = f"{diff}"
            else:
                diff_str = np.array_str(diff, max_line_width=120, precision=12, suppress_small=False)
                diff_str = "\n".join("    " + ln for ln in diff_str.splitlines())

        if xptd.shape == ():
            message = """\t{}: computed value ({}) does not match ({}) by difference ({}).""".format(
                label, cptd_str, xptd_str, diff_str
            )
        else:
            message = """\t{}: computed value does not match.\n  Expected:\n{}\n  Observed:\n{}\n  Difference:\n{}\n""".format(
                label, xptd_str, cptd_str, diff_str
            )

    return return_handler(allclose, label, message, return_message, quiet)


def _compare_recursive(expected, computed, atol, rtol, _prefix=False, equal_phase=False):

    errors = []
    name = _prefix or "root"
    prefix = name + "."

    # Initial conversions if required
    if isinstance(expected, BaseModel):
        expected = expected.dict()

    if isinstance(computed, BaseModel):
        computed = computed.dict()

    if isinstance(expected, (str, int, bool, complex)):
        if expected != computed:
            errors.append((name, "Value {} did not match {}.".format(expected, computed)))

    elif isinstance(expected, (list, tuple)):
        try:
            if len(expected) != len(computed):
                errors.append((name, "Iterable lengths did not match"))
            else:
                for i, item1, item2 in zip(range(len(expected)), expected, computed):
                    errors.extend(
                        _compare_recursive(
                            item1, item2, _prefix=prefix + str(i), atol=atol, rtol=rtol, equal_phase=equal_phase
                        )
                    )
        except TypeError:
            errors.append((name, "Expected computed to have a __len__()"))

    elif isinstance(expected, dict):
        expected_extra = computed.keys() - expected.keys()
        computed_extra = expected.keys() - computed.keys()
        if len(expected_extra):
            errors.append((name, "Found extra keys {}".format(expected_extra)))
        if len(computed_extra):
            errors.append((name, "Missing keys {}".format(computed_extra)))

        for k in expected.keys() & computed.keys():
            name = prefix + str(k)
            errors.extend(
                _compare_recursive(
                    expected[k], computed[k], _prefix=name, atol=atol, rtol=rtol, equal_phase=equal_phase
                )
            )

    elif isinstance(expected, (float, np.number)):
        passfail, msg = compare_values(
            expected, computed, atol=atol, rtol=rtol, equal_phase=equal_phase, return_message=True, quiet=True
        )
        if not passfail:
            errors.append((name, "Arrays differ." + msg))

    elif isinstance(expected, np.ndarray):
        if np.issubdtype(expected.dtype, np.floating):
            passfail, msg = compare_values(
                expected, computed, atol=atol, rtol=rtol, equal_phase=equal_phase, return_message=True, quiet=True
            )
        else:
            passfail, msg = compare(expected, computed, equal_phase=equal_phase, return_message=True, quiet=True)
        if not passfail:
            errors.append((name, "Arrays differ." + msg))

    elif isinstance(expected, type(None)):
        if expected is not computed:
            errors.append((name, "'None' does not match."))

    else:
        errors.append((name, f"Type {type(expected)} not understood -- stopping recursive compare."))

    return errors


def compare_recursive(
    expected: Union[Dict, BaseModel, "ProtoModel"],  # type: ignore
    computed: Union[Dict, BaseModel, "ProtoModel"],  # type: ignore
    label: str = None,
    *,
    atol: float = 1.0e-6,
    rtol: float = 1.0e-16,
    forgive: List[str] = None,
    equal_phase: Union[bool, List] = False,
    quiet: bool = False,
    return_message: bool = False,
    return_handler: Callable = None,
) -> bool:
    """
    Recursively compares nested structures such as dictionaries and lists.

    Parameters
    ----------
    expected : dict
        Reference value against which `computed` is compared.
        Dict may be of any depth but should contain Plain Old Data.
    computed : int, bool, str or int array-like
        Input value to compare against `expected`.
        Dict may be of any depth but should contain Plain Old Data.
    atol : int or float, optional
        Absolute tolerance (see formula below).
    label : str, optional
        Label for passed and error messages. Defaults to calling function name.
    rtol : float, optional
        Relative tolerance (see formula below). By default set to zero so `atol` dominates.
    forgive : list, optional
        Keys in top level which may change between `expected` and `computed` without triggering failure.
    equal_phase : bool, optional
        Compare computed *or its opposite* as equal.

    Returns
    -------
    allclose : bool
        Returns True if `expected` and `computed` are equal within tolerance; False otherwise.
    message : str, optional
        When return_message=True, also return passed or error message.

    Notes
    -----

    .. code-block:: python

        absolute(computed - expected) <= (atol + rtol * absolute(expected))

    """
    label = label or sys._getframe().f_back.f_code.co_name
    if atol >= 1:
        raise ValueError(
            "Prior to v0.4.0, ``compare_recursive`` used to 10**-atol any atol >=1. That has ceased, so please express your atol literally."
        )
    if return_handler is None:
        return_handler = _handle_return

    errors = _compare_recursive(expected, computed, atol=atol, rtol=rtol)

    if errors and equal_phase:
        n_errors = _compare_recursive(expected, computed, atol=atol, rtol=rtol, equal_phase=True)
        n_errors = dict(n_errors)

        if equal_phase is False:
            equal_phase = []
        elif equal_phase is True:
            equal_phase = list(dict(errors).keys())
        else:
            equal_phase = [(ep if ep.startswith("root.") else "root." + ep) for ep in equal_phase]
        phased = []

        for nomatch in sorted(errors):
            for ep in equal_phase or []:
                if nomatch[0].startswith(ep):
                    if nomatch[0] not in n_errors:
                        phased.append(nomatch)
                        errors.remove(nomatch)

    if forgive is None:
        forgive = []
    else:
        forgive = [(fg if fg.startswith("root.") else "root." + fg) for fg in forgive]
    forgiven = []

    for nomatch in sorted(errors):
        for fg in forgive or []:
            if nomatch[0].startswith(fg):
                forgiven.append(nomatch)
                errors.remove(nomatch)

    ## print if verbose >= 2 if these functions had that knob
    # forgiven_message = []
    # for e in sorted(forgiven):
    #     forgiven_message.append(e[0])
    #     forgiven_message.append("forgiven    " + e[1])
    # pprint.pprint(forgiven)

    message = []
    for e in sorted(errors):
        message.append(e[0])
        message.append("    " + e[1])

    ret_msg_str = "\n".join(message)

    return return_handler(len(ret_msg_str) == 0, label, ret_msg_str, return_message, quiet)


def compare_molrecs(
    expected,
    computed,
    label: str = None,
    *,
    atol: float = 1.0e-6,
    rtol: float = 1.0e-16,
    forgive=None,
    verbose: int = 1,
    relative_geoms="exact",
    return_message: bool = False,
    return_handler: Callable = None,
) -> bool:
    """Function to compare Molecule dictionaries. Prints
    #    :py:func:`util.success` when elements of `computed` match elements of
    #    `expected` to `tol` number of digits (for float arrays).

    """
    # Need to manipulate the dictionaries a bit, so hold values
    xptd = copy.deepcopy(expected)
    cptd = copy.deepcopy(computed)

    def massage_dicts(dicary):
        # if 'fix_symmetry' in dicary:
        #     dicary['fix_symmetry'] = str(dicary['fix_symmetry'])
        # if 'units' in dicary:
        #     dicary['units'] = str(dicary['units'])
        if "fragment_files" in dicary:
            dicary["fragment_files"] = [str(f) for f in dicary["fragment_files"]]
        # and about int vs long errors
        # if 'molecular_multiplicity' in dicary:
        #     dicary['molecular_multiplicity'] = int(dicary['molecular_multiplicity'])
        # if 'fragment_multiplicities' in dicary:
        #     dicary['fragment_multiplicities'] = [(m if m is None else int(m))
        #                                          for m in dicary['fragment_multiplicities']]
        if "fragment_separators" in dicary:
            dicary["fragment_separators"] = [(s if s is None else int(s)) for s in dicary["fragment_separators"]]
        # forgive generator version changes
        if "provenance" in dicary:
            dicary["provenance"].pop("version")
        # regularize connectivity ordering
        if "connectivity" in dicary:
            conn = [(min(at1, at2), max(at1, at2), bo) for (at1, at2, bo) in dicary["connectivity"]]
            conn.sort(key=lambda tup: tup[0])
            dicary["connectivity"] = conn

        return dicary

    xptd = massage_dicts(xptd)
    cptd = massage_dicts(cptd)

    if relative_geoms == "exact":
        pass
    elif relative_geoms == "align":
        # can't just expect geometries to match, so we'll align them, check that
        #   they overlap and that the translation/rotation arrays jibe with
        #   fix_com/orientation, then attach the oriented geom to computed before the
        #   recursive dict comparison.
        from .molutil.align import B787

        cgeom = np.array(cptd["geom"]).reshape((-1, 3))
        rgeom = np.array(xptd["geom"]).reshape((-1, 3))
        rmsd, mill = B787(
            rgeom=rgeom,
            cgeom=cgeom,
            runiq=None,
            cuniq=None,
            atoms_map=True,
            mols_align=True,
            run_mirror=False,
            verbose=0,
        )
        if cptd["fix_com"]:
            return compare(
                True,
                np.allclose(np.zeros((3)), mill.shift, atol=atol),
                "null shift",
                quiet=(verbose == 0),
                return_message=return_message,
                return_handler=return_handler,
            )
        if cptd["fix_orientation"]:
            return compare(
                True,
                np.allclose(np.identity(3), mill.rotation, atol=atol),
                "null rotation",
                quiet=(verbose == 0),
                return_message=return_message,
                return_handler=return_handler,
            )
        ageom = mill.align_coordinates(cgeom)
        cptd["geom"] = ageom.reshape((-1))

    return compare_recursive(
        xptd,
        cptd,
        atol=atol,
        rtol=rtol,
        label=label,
        forgive=forgive,
        quiet=(verbose == 0),
        return_message=return_message,
        return_handler=return_handler,
    )



import collections
import re
from typing import List


def order_molecular_formula(formula: str, order: str = "alphabetical") -> str:
    """
    Reorders a molecular formula.

    Parameters
    ----------
    formula: str
        A molecular formula
    order: str, optional
        Sorting order of the formula. Valid choices are "alphabetical" and "hill".

    Returns
    -------
    str
        The molecular formula.
    """

    matches = re.findall(r"[A-Z][^A-Z]*", formula)
    if not "".join(matches) == formula:
        raise ValueError(f"{formula} is not a valid molecular formula.")
    count = collections.defaultdict(int)
    for match in matches:
        match_n = re.match(r"(\D+)(\d*)", match)
        assert match_n
        if match_n.group(2) == "":
            n = 1
        else:
            n = int(match_n.group(2))
        count[match_n.group(1)] += n
    symbols = [k for k, v in count.items() for i in range(v)]
    return molecular_formula_from_symbols(symbols=symbols, order=order)


def molecular_formula_from_symbols(symbols: List[str], order: str = "alphabetical") -> str:
    """
    Returns the molecular formula for a list of symbols.

    Parameters
    ----------
    symbols: List[str]
        List of chemical symbols
    order: str, optional
        Sorting order of the formula. Valid choices are "alphabetical" and "hill".

    Returns
    -------
    str
        The molecular formula.
    """

    supported_orders = ["alphabetical", "hill"]
    order = order.lower()
    if order not in supported_orders:
        raise ValueError(f"Unsupported molecular formula order: {order}. Supported orders are f{supported_orders}.")
    count = collections.Counter(x.title() for x in symbols)
    element_order = sorted(count.keys())

    if order == "hill" and "C" in element_order:
        if "H" in element_order:
            element_order.insert(0, element_order.pop(element_order.index("H")))
        element_order.insert(0, element_order.pop(element_order.index("C")))

    ret = []
    for k in element_order:
        c = count[k]
        ret.append(k)
        if c > 1:
            ret.append(str(c))

    return "".join(ret)
